# Copyright (c) 2014-2017 Louisiana State University
#               2017-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Input/output methods for tabular data.

Access to HACR triggers requires local access to the MySQL database. Users
can set the IP address of the server, and the username and password for
connections in the following environment variables

- ``HACR_DATABASE_SERVER``
- ``HACR_DATABASE_USER``
- ``HACR_DATABASE_PASSWD``

These can also be given directly to the connection function as keyword
arguments
"""

from __future__ import annotations

import os.path
from functools import (
    cache,
    partial,
)
from typing import (
    TYPE_CHECKING,
    cast,
)

from astropy.table import (
    Column,
    Table,
    vstack,
)
from dateutil.relativedelta import relativedelta

from ...segments import Segment
from ...time import (
    from_gps,
    to_gps,
)
from .. import EventTable
from . import sql as io_sql
from .utils import dynamic_columns

if TYPE_CHECKING:
    from collections.abc import (
        Callable,
        Iterable,
        Iterator,
        Mapping,
        Sequence,
    )

    import sqlalchemy

    from ...time import SupportsToGps
    from .sql import WhereExpression

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"

# get HACR environment variables
HACR_DATABASE_SERVER = os.getenv("HACR_DATABASE_SERVER", None)
HACR_DATABASE_USER = os.getenv("HACR_DATABASE_USER", None)
HACR_DATABASE_QUERY = {
    "charset": "utf8",
}


# -- dynamic columns -----------------

def gpstime(table: Table, name: str = "gpstime") -> Column:
    """Combine the ``gps_start`` and ``gps_offset`` columns."""
    return Column(
        table["gps_start"][:] + table["gps_offset"][:],
        name=name,
    )


DYNAMIC_COLUMN_FUNC: dict[str, Callable] = {
    "gps": partial(gpstime, name="gps"),
    "gpstime": gpstime,
    "time": gpstime,
}
DYNAMIC_COLUMN_INPUT: dict[str, set[str]] = {
    "gps": (_gps_columns := {"gps_start", "gps_offset"}),
    "gpstime": _gps_columns,
    "time": _gps_columns,
}


# -- HACR DB integration -------------

def get_database_names(
    start: SupportsToGps,
    end: SupportsToGps,
) -> list[str]:
    """Return the list of HACR database names covering the given interval.

    Parameters
    ----------
    start : `~gwpy.time.LIGOTimeGPS`, `float`, `str`, optional
        GPS start time of required data, defaults to start of data found;
        any input parseable by `~gwpy.time.to_gps` is fine.

    end : `~gwpy.time.LIGOTimeGPS`, `float`, `str`, optional
        GPS end time of required data, defaults to end of data found;
        any input parseable by `~gwpy.time.to_gps` is fine.

    Returns
    -------
    databases : `list` of `str`
        The list of database names that represent the semi-open interval
        ``[start, end)``.

    Examples
    --------
    >>> get_database_names('Jan 1 2024', 'Mar 1 2024')
    ['geo202401', 'geo202402']
    """
    start = int(to_gps(start))
    end = int(to_gps(end))
    if end < start:
        msg = "invalid start/end for HACR query, end must be greater than start"
        raise ValueError(msg)
    if start == end:
        end += 1
    # convert to datetimes
    d = from_gps(start)
    enddate = from_gps(end)
    # loop over months
    dbs = []
    dt = relativedelta(months=1)
    while d < enddate:
        dbs.append(f"geo{d.strftime('%Y%m')}")
        d += dt
    return dbs


def get_hacr_channels(
    gps: SupportsToGps | None = None,
    database: str | None = None,
    engine: sqlalchemy.Engine | None = None,
    **kwargs,
) -> list[str]:
    """Return the names of all channels present in the given HACR database.

    Parameters
    ----------
    gps : `~gwpy.time.LIGOTimeGPS`, `float`, `str`, optional
        GPS time of data query, defaults to ```"now"``.
        Any input parseable by `~gwpy.time.to_gps` is fine.

    database : `str`, optional
        The name of the database to query.
        Required if ``engine`` is not given.

    engine : `sqlalchemy.engine.Engine`, optional
        An existing database connection engine.
        If not given, one will be created using the other connection parameters.

    kwargs
        All other keyword arguments are passed to `create_engine`.

    Returns
    -------
    channels : `list` of `str`
        The list of channel names present in the database.

    See Also
    --------
    gwpy.table.io.hacr.create_engine
        For details of valid keyword arguments to create the
        `sqlalchemy.Engine`.
    """
    if engine is None:
        if gps is None:
            gps = to_gps("now") - 43200  # 12 hours ago
        if database is None:
            database, = get_database_names(gps, gps)
        engine = create_engine(
            database,
            **kwargs,
        )
    # query
    with engine.connect() as conn:
        result = conn.execute(
            io_sql.format_query(
                "job",
                columns=["channel"],
                where=["monitorName == 'chacr'"],
            ),
        )
        return list({r.channel for r in result})


def get_hacr_triggers(
    # query options
    channel: str | None = None,
    start: SupportsToGps | None = None,
    end: SupportsToGps | None = None,
    columns: Iterable[str] | None = None,
    tablename: str = "mhacr",
    process_id: int | None = None,
    monitor: str = "chacr",
    where: WhereExpression | Iterable[WhereExpression] = None,
    # connection options
    engine: sqlalchemy.Engine | None = None,
    drivername: str = "mysql+pymysql",
    username: str | None = HACR_DATABASE_USER,
    password: str | None = None,
    host: str | None = HACR_DATABASE_SERVER,
    port: int | None = None,
    database: str | None = None,
    query: Mapping[str, Sequence[str] | str] | None = None,
    **kwargs,
) -> Table:
    """Fetch a table of HACR triggers in the given interval.

    Parameters
    ----------
    channel : `str`
        The name of the data channel for which to query.

    start : `~gwpy.time.LIGOTimeGPS`, `float`, `str`, optional
        GPS start time of required data, defaults to start of data found;
        any input parseable by `~gwpy.time.to_gps` is fine.

    end : `~gwpy.time.LIGOTimeGPS`, `float`, `str`, optional
        GPS end time of required data, defaults to end of data found;
        any input parseable by `~gwpy.time.to_gps` is fine.

    columns : `list` of `str`, optional
        The columns to select.
        Default is all columns.

    tablename : `str`, optional
        The name of the database table to select data from.
        Default is ``"mhacr"``.

    process_id : `int`, optional
        The HACR ``process_id`` to match.
        Default is to not select based on ``process_id``.

    monitor : `str`, optional
        The name of the database table to select.
        Default is ``"chacr"``.

    where : `sqlalchemy.sql.elements.SQLColumnExpression`, optional
        A SQLAlchemy expression to use as the ``WHERE`` condition in
        the query.
        Default is `None`.

    engine : `sqlalchemy.engine.Engine`, optional
        An existing database connection engine.
        If not given, one will be created using the other connection parameters.

    drivername : `str`, optional
        The database backend and driver name.
        Default is ``"mysql+pymysql"``.

    username : `str`, optional
        The username for authentication to the database.
        Defaults to the value of the ``HACR_DATABASE_USER`` environment variable.

    password : `str`, optional
        The password for authentication to the database.
        Defaults to the value of the ``HACR_DATABASE_PASSWD`` environment variable.

    host : `str`, optional
        The hostname of the database server.
        Defaults to the value of the ``HACR_DATABASE_SERVER`` environment variable.

    port : `int`, optional
        The port number of the database server.
        Default is the default port for the given ``drivername``.

    database : `str`, optional
        The name of the database to query.
        Required if ``engine`` is not given.

    query : `dict`, optional
        Additional query parameters to use in the database connection.

    kwargs
        All other keyword arguments are passed to `pandas.read_sql`.

    Returns
    -------
    table : `gwpy.table.EventTable`
        A table containing the HACR triggers received from the database.
    """
    # parse given where
    where = [io_sql.format_where(where)]

    # format GPS interval
    if start is not None:
        start = to_gps(start)
    if end is not None:
        end = to_gps(end)

    # get database names to query
    databases: Iterable[str | None]
    if engine is None and database is None:
        if start is None or end is None:
            msg = (
                "start and end options must be given to determine "
                "databases to query"
            )
            raise ValueError(msg)
        databases = get_database_names(start, end)
    else:
        databases = [database]

    # initialise column name holders
    valid_columns: dict[str, type] | None = None
    read_cols: set[str] | None = None
    dynamic_cols: set[str] | None = None

    # loop over databases
    tables: list[EventTable] = []
    for db in databases:
        if engine is None:
            engine = create_engine(
                db,  # type: ignore[arg-type]
                drivername=drivername,
                username=username,
                password=password,
                host=host,
                port=port,
                query=query,
            )

        # format columns
        if valid_columns is None:  # only once
            valid_columns = io_sql.get_columns(
                tablename,
                engine,
            )
            read_cols, dynamic_cols = dynamic_columns(
                columns,
                valid_columns,
                DYNAMIC_COLUMN_INPUT,
            )

        # loop over process_id
        for block in _query_database(
            engine,
            start=start,
            end=end,
            channel=channel,
            columns=read_cols,
            tablename=tablename,
            process_tablename="job",
            monitor=monitor,
            process_id=process_id,
            where=where,
            **kwargs,
        ):
            if len(block):
                tables.append(block)

    if tables:  # join tables we got
        table = vstack(
            tables,
            join_type="exact",
            metadata_conflicts="error",
        )
    elif valid_columns is not None:  # create empty table
        names, dtypes = zip(*(
            (name, valid_columns.get(name))
            for name in read_cols or []
        ), strict=True)
        table = Table(
            names=names,
            dtype=dtypes,
        )
    else:  # no databases found
        msg = "no HACR data found from query"
        raise RuntimeError(msg)

    # add dynamic columns
    if dynamic_cols:
        # if dynamic cols is populated,
        # then columns and read_cols _must_ also be populated
        columns = cast("Iterable[str]", columns)
        read_cols = cast("set[str]", read_cols)
        # generate requested derived columns on-the-fly
        for col_name in dynamic_cols:
            col_data = DYNAMIC_COLUMN_FUNC[col_name](table)
            table.add_column(col_data, name=col_name)
        # remove columns that were only added to generate a derived column
        for col_name in read_cols - set(columns):
            table.remove_column(col_name)

    return table


@cache
def _process_id_query(
    tablename: str,
    monitor: str,
    channel: str | None = None,
) -> sqlalchemy.Select:
    """Construct the ``SELECT`` query to get valid HACR ``process_id``s."""
    where = [
        f"monitorName == '{monitor}'",
    ]
    if channel is not None:
        where.append(f"channel == '{channel}'")
    return io_sql.format_query(
        tablename,
        columns=[
            "process_id",
            "gps_start",
            "gps_stop",
        ],
        where=where,
    )


def _query_database(
    engine: sqlalchemy.Engine,
    start: SupportsToGps | None,
    end: SupportsToGps | None,
    channel: str | None,
    process_id: int | None,
    columns: Iterable[str] | None,
    tablename: str,
    monitor: str,
    process_tablename: str,
    where: Iterable[WhereExpression],
    **kwargs,
) -> Iterator[EventTable]:
    """Query a database for HACR triggers.

    Yields `astropy.table.Table` instances for each ``process_id``.
    """
    # allow querying for all values
    if start is None:
        start = 0
    if end is None:
        end = 1e10

    with engine.connect() as conn:
        if process_id is None:
            # find process ID(s) for this channel
            pids = [
                (row.process_id, row.gps_start, row.gps_stop)
                for row in conn.execute(_process_id_query(
                    process_tablename,
                    monitor,
                    channel=channel,
                ))
            ]
        else:
            pids = [(process_id, start, end)]

        for pid, pid_start, pid_end in pids:
            try:   # get overlap between the given range and the process
                span = Segment(start, end) & Segment(pid_start, pid_end)
            except ValueError:  # no overlap, carry on
                continue
            yield _query_hacr(
                engine,
                start=span[0],
                end=span[1],
                process_id=pid,
                tablename=tablename,
                columns=columns,
                where=where,
                **kwargs,
            )


def _query_hacr(
    engine: sqlalchemy.Engine,
    start: SupportsToGps | None,
    end: SupportsToGps | None,
    process_id: int | None,
    tablename: str,
    columns: Iterable[str] | None,
    where: Iterable[WhereExpression],
    **kwargs,
) -> EventTable:
    """Query for HACR triggers matching the given process_id.

    Returns
    -------
    table : `astropy.table.Table`
        The table of triggers received from the database.
    """
    # recast to append without impacting variable outside scope
    where = list(where)
    # add conditions for this request
    if start is not None:
        where.append(f"gps_start >= {start}")
    if end is not None:
        where.append(f"gps_start < {end}")
    if process_id is not None:
        where.append(f"process_id == {process_id}")

    return EventTable.fetch(
        source="sql",
        engine=engine,
        tablename=tablename,
        columns=columns,
        where=where,
        order_by="gps_start",
        **kwargs,
    )


EventTable.fetch.registry.register_reader(
    "hacr",
    EventTable,
    get_hacr_triggers,
)


# -- utilities -----------------------

def create_engine(
    database: str,
    drivername: str = "mysql+pymysql",
    host: str | None = HACR_DATABASE_SERVER,
    username: str | None = HACR_DATABASE_USER,
    password: str | None = None,
    query: Mapping[str, Sequence[str] | str] | None = None,
    **kwargs,
) -> sqlalchemy.Engine:
    """Create an `sqlalchemy.engine.Engine` for HACR.

    This is just a thin wrapper around `gwpy.table.io.sql.create_engine` with
    HACR defaults.
    """
    if password is None:
        password = os.getenv("HACR_DATABASE_PASSWD") or None
    if query is None:
        query = HACR_DATABASE_QUERY
    return io_sql.create_engine(
        drivername,
        host=host,
        username=username,
        password=password,
        database=database,
        query=query,
        **kwargs,
    )



# Copyright (c) 2014-2017 Louisiana State University
#               2017-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Unit test for timeseries module."""

import os
from io import BytesIO

import numpy
import pytest
from astropy import units
from matplotlib import rc_context

from ...detector import Channel
from ...segments import Segment
from ...testing import mocks, utils
from ...testing.errors import pytest_skip_flaky_network
from ...time import LIGOTimeGPS, Time
from ...types import Array2D
from .. import (
    Bits,
    StateTimeSeries,
    StateTimeSeriesDict,
    StateVector,
    StateVectorDict,
    StateVectorList,
)
from .test_core import (
    TestTimeSeriesBase as _TestTimeSeriesBase,
    TestTimeSeriesBaseDict as _TestTimeSeriesBaseDict,
    TestTimeSeriesBaseList as _TestTimeSeriesBaseList,
)
from .test_timeseries import (
    GWOSC_GW150914_DQ_BITS,
    GWOSC_GW150914_DQ_NAME,
    # Description of GW150914
    GWOSC_GW150914_IFO,
    GWOSC_GW150914_INJ_BITS,
    GWOSC_GW150914_INJ_NAME,
    GWOSC_GW150914_SEGMENT,
    GWOSC_GW190814_DQ_BITS,
    GWOSC_GW190814_DQ_NAME,
    # Description of GW190814
    GWOSC_GW190814_IFO,
    GWOSC_GW190814_INJ_BITS,
    GWOSC_GW190814_INJ_NAME,
    GWOSC_GW190814_SEGMENT,
)

# This is used to test reading of injections
GWOSC_GW150914_INJ_KWARGS = {
    "hdf5": {
        "path": "quality/injections",
        "value_dataset": "Injmask",
        "bits_dataset": "InjDescriptions",
    },
    "gwf": {
        "channel": GWOSC_GW150914_INJ_NAME["gwf"],
    },
}

GWOSC_GW190814_INJ_KWARGS = {
    "hdf5": {
        "path": "quality/injections",
        "value_dataset": "Injmask",
        "bits_dataset": "InjDescriptions",
    },
    "gwf": {
        "channel": GWOSC_GW190814_INJ_NAME["gwf"],
    },
}

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"


# -- StateTimeSeries ----------------------------------------------------------

class TestStateTimeSeries(_TestTimeSeriesBase):
    TEST_CLASS = StateTimeSeries

    @classmethod
    def setup_class(cls):
        cls.data = numpy.asarray(
            [0, 1, 1, 1, 0, 0, 0, 1, 0, 0, 1, 1, 1, 0, 1, 0, 1, 1, 1, 1],
            dtype=bool,
        )

    def test_new(self):
        sts = self.create()
        assert isinstance(sts, self.TEST_CLASS)
        assert sts.dtype is numpy.dtype("bool")

    def test_getitem(self, array):
        assert isinstance(array[0], numpy.bool_)

    def test_unit(self, array):
        assert array.unit is units.dimensionless_unscaled

        # check that we can't delete the unit
        with pytest.raises(AttributeError):
            del array.unit

        # check that we can't set the unit
        with pytest.raises(TypeError):
            self.create(unit="test")

    def test_math(self, array):
        # test that math operations give back booleans
        a2 = array ** 2
        assert a2.dtype is numpy.dtype(bool)
        utils.assert_array_equal(array.value, a2.value)

    @pytest.mark.requires("arrakis")
    @pytest.mark.parametrize("copy", [False, True])
    @pytest.mark.parametrize("dtype", [
        pytest.param(bool, id="bool"),
        pytest.param(numpy.dtype('bool'), id="np_bool"),
    ])
    def test_from_arrakis(self, copy, dtype):
        """Test `StateTimeSeries.from_arrakis`."""
        super().test_from_arrakis(copy, dtype)

    @pytest.mark.requires("nds2")
    def test_from_nds2_buffer(self):
        # build fake buffer
        nds_buffer = mocks.nds2_buffer(
            "X1:TEST",
            self.data,
            1000000000,
            self.data.shape[0],
            "m",
            name="test",
        )
        array = self.TEST_CLASS.from_nds2_buffer(nds_buffer)
        assert array.unit is units.dimensionless_unscaled
        assert array.dtype is numpy.dtype(bool)

    def test_to_dqflag(self, array):
        flag = array.to_dqflag()
        utils.assert_segmentlist_equal(
            flag.active,
            map(Segment, ((1, 4), (7, 8), (10, 13), (14, 15), (16, 20))),
        )
        utils.assert_segmentlist_equal(flag.known, [array.span])
        assert flag.name == array.name
        assert flag.label == array.name
        assert flag.description is None

        flag = array.to_dqflag(
            minlen=2,
            dtype=LIGOTimeGPS,
            name="Test flag",
            round=True,
            label="Label",
            description="Description",
        )
        utils.assert_segmentlist_equal(
            flag.active,
            map(Segment, ((1, 4), (10, 13), (16, 20))),
        )
        assert isinstance(flag.active[0][0], LIGOTimeGPS)
        assert flag.name == "Test flag"
        assert flag.label == "Label"
        assert flag.description == "Description"

    def test_override_unit(self, array):
        with pytest.raises(NotImplementedError):
            super().test_override_unit(array)

    def test_is_compatible_error_unit(self, array):
        pytest.skip(f"not implemented for {self.TEST_CLASS.__name__}")

    def test_to_from_pycbc(self):
        pytest.skip(f"not implemented for {self.TEST_CLASS.__name__}")

    def test_to_from_lal(self, array):
        with pytest.raises(NotImplementedError):
            super().test_to_from_lal(array)

    @pytest.mark.parametrize("copy", (False, True))
    def test_to_from_lal_no_copy(self, array, copy):
        with pytest.raises(NotImplementedError):
            super().test_to_from_lal_no_copy(array, copy)

    @pytest.mark.requires("lal")
    def test_to_from_lal_pow10_units(self, array):
        with pytest.raises(NotImplementedError):
            super().test_to_from_lal_pow10_units(array)

    @pytest.mark.requires("lal")
    def test_to_from_lal_scaled_units(self, array):
        with pytest.raises(NotImplementedError):
            super().test_to_from_lal_scaled_units(array)

    @pytest.mark.requires("lal")
    def test_to_from_lal_unrecognised_units(self, array):
        with pytest.raises(NotImplementedError):
            super().test_to_from_lal_unrecognised_units(array)


# -- StateTimeSeriesDict ------------------------------------------------------

class TestStateTimeSeriesDict(_TestTimeSeriesBaseDict):
    TEST_CLASS = StateTimeSeriesDict
    ENTRY_CLASS = StateTimeSeries
    DTYPE = bool

    def test_resample(self, instance):
        pytest.skip(f"not implemented for {type(instance).__name__}")


# -- Bits ---------------------------------------------------------------------

class TestBits:
    TEST_CLASS = Bits

    @pytest.mark.parametrize(("in_", "out"), [
        # list
        pytest.param(
            ["bit 0", "bit 1", "bit 2", None, "bit 3", ""],
            ["bit 0", "bit 1", "bit 2", None, "bit 3", None],
            id="list",
        ),
        # dict
        pytest.param(
            {1: "bit 1", 4: "bit 4", "6": "bit 6"},
            [None, "bit 1", None, None, "bit 4", None, "bit 6"],
            id="dict",
        ),
    ])
    def test_init(self, in_, out):
        """Test initialising `Bits`."""
        bits = self.TEST_CLASS(in_)
        assert bits == out
        assert bits.channel is None
        assert bits.epoch is None
        assert bits.description == {bit: None for bit in bits if bit is not None}

        bits = self.TEST_CLASS(in_, channel="L1:Test", epoch=0)
        assert bits.epoch == Time(0, format="gps")
        assert bits.channel == Channel("L1:Test")

    def test_str(self):
        """Test ``str(Bits)``."""
        bits = self.TEST_CLASS(["a", "b", None, "c"])
        assert str(bits) == os.linesep.join([
            "Bits(0: a",
            "     1: b",
            "     3: c,",
            "     channel=None,",
            "     epoch=None)",
        ])

    def test_repr(self):
        """Test ``repr(Bits)``."""
        bits = self.TEST_CLASS(["a", "b", None, "c"])
        assert repr(bits) == os.linesep.join([
            "<Bits(0: 'a'",
            "      1: 'b'",
            "      3: 'c',",
            "      channel=None,",
            "      epoch=None)>",
        ])

    def test_array(self):
        """Test converting `Bits` to an array."""
        bits = self.TEST_CLASS(["a", "b", None, "c"])
        utils.assert_array_equal(
            numpy.asarray(bits),
            ["a", "b", "", "c"],
        )


# -- StateVector---------------------------------------------------------------

class TestStateVector(_TestTimeSeriesBase):
    TEST_CLASS = StateVector
    DTYPE = "uint32"

    @classmethod
    def setup_class(cls):
        numpy.random.seed(0)
        cls.data = numpy.random.randint(
            2**4+1,
            size=100,
        ).astype(cls.DTYPE, copy=False)

    def test_bits(self, array):
        assert isinstance(array.bits, Bits)
        assert array.bits == [f"Bit {i}" for i in range(32)]

        bits = [f"Bit {i}" for i in range(4)]

        sv = self.create(bits=bits)
        assert isinstance(sv.bits, Bits)
        assert sv.bits.channel is sv.channel
        assert sv.bits.epoch == sv.epoch
        assert sv.bits == bits

        del sv.bits
        del sv.bits
        assert isinstance(sv.bits, Bits)
        assert sv.bits == [f"Bit {i}" for i in range(32)]

        sv = self.create(dtype="uint16")
        assert sv.bits == [f"Bit {i}" for i in range(16)]

    def test_boolean(self, array):
        b = array.boolean
        assert isinstance(b, Array2D)
        assert b.shape == (array.size, len(array.bits))
        # array[0] == 12, check boolean equivalent
        utils.assert_array_equal(b[0], [12 >> j & 1 for j in range(32)])

    def test_get_bit_series(self, array):
        # test default
        bs = array.get_bit_series()
        assert isinstance(bs, StateTimeSeriesDict)
        assert list(bs.keys()) == array.bits
        # check that correct number of samples match (simple test)
        assert bs["Bit 2"].sum() == 43

        # check that bits in gives bits out
        bs = array.get_bit_series(["Bit 0", "Bit 3"])
        assert list(bs.keys()) == ["Bit 0", "Bit 3"]
        assert [v.sum() for v in bs.values()] == [50, 41]

        # check that invalid bits throws exception
        with pytest.raises(
            ValueError,
            match=r"^Bit 'blah' not found in StateVector$",
        ):
            array.get_bit_series(["blah"])

    def test_plot(self, array):
        with rc_context(rc={"text.usetex": False}):
            plot = array.plot()
            # make sure there were no lines drawn
            assert len(plot.gca().lines) == 0
            # assert one collection for each of known and active segmentlists
            assert len(plot.gca().collections) == len(array.bits) * 2
            plot.save(BytesIO(), format="png")
            plot.close()

            # test timeseries plotting as normal
            plot = array.plot(format="timeseries")
            line = plot.gca().lines[0]
            utils.assert_array_equal(line.get_xdata(), array.xindex.value)
            utils.assert_array_equal(line.get_ydata(), array.value)
            plot.close()

    def test_resample(self, array):
        # check downsampling by factor of 2
        a2 = array.resample(array.sample_rate / 2.)
        assert a2.sample_rate == array.sample_rate / 2.
        assert a2.bits is array.bits
        utils.assert_array_equal(
            a2.value[:10],
            [12, 0, 3, 0, 4, 0, 6, 5, 8, 0],
        )

        # check upsampling raises NotImplementedError
        with pytest.raises(NotImplementedError):
            array.resample(array.sample_rate * 2.)

        # check resampling by non-integer factor raises error
        with pytest.raises(ValueError):
            array.resample(array.sample_rate * .75)
        with pytest.raises(ValueError):
            array.resample(array.sample_rate * 1.5)

    def test_to_from_lal_scaled_units(self):
        pytest.skip(
            f"not implemented for {self.TEST_CLASS.__name__}",
        )

    # -- data access ----------------------------

    @pytest.mark.parametrize("format", [
        "hdf5",
        pytest.param(  # only frameCPP actually reads units properly
            "gwf",
            marks=pytest.mark.requires("LDAStools.frameCPP"),
        ),
    ])
    @pytest_skip_flaky_network
    def test_fetch_open_data(self, format):
        sv = self.TEST_CLASS.fetch_open_data(
            GWOSC_GW150914_IFO,
            GWOSC_GW150914_SEGMENT[0],
            GWOSC_GW150914_SEGMENT[1],
            format=format,
            version=3,
        )
        ref = StateVector(
            [127, 127, 127, 127],
            unit="",
            t0=GWOSC_GW150914_SEGMENT[0],
            dt=1,
            name=GWOSC_GW150914_DQ_NAME[format],
            bits=GWOSC_GW150914_DQ_BITS[format],
        )
        utils.assert_quantity_sub_equal(sv, ref, exclude=["channel", "bits"])
        assert sorted(map(str, sv.bits)) == sorted(map(str, ref.bits))

    @pytest.mark.parametrize("format", [
        "hdf5",
        pytest.param(  # only frameCPP actually reads units properly
            "gwf",
            marks=pytest.mark.requires("LDAStools.frameCPP"),
        ),
    ])
    @pytest_skip_flaky_network
    def test_fetch_open_data_new_format(self, format):
        sv = self.TEST_CLASS.fetch_open_data(
            GWOSC_GW190814_IFO,
            GWOSC_GW190814_SEGMENT[0],
            GWOSC_GW190814_SEGMENT[1],
            format=format,
            version=3,
        )
        ref = StateVector(
            [127, 127, 127, 127],
            unit="",
            t0=GWOSC_GW190814_SEGMENT[0],
            dt=1,
            name=GWOSC_GW190814_DQ_NAME[format],
            bits=GWOSC_GW190814_DQ_BITS[format],
        )
        utils.assert_quantity_sub_equal(sv, ref, exclude=["channel", "bits"])
        assert sorted(map(str, sv.bits)) == sorted(map(str, ref.bits))


    @pytest.mark.parametrize("format", [
        "hdf5",
        pytest.param(  # only frameCPP actually reads units properly
            "gwf",
            marks=pytest.mark.requires("LDAStools.frameCPP"),
        ),
    ])
    @pytest_skip_flaky_network
    def test_fetch_open_data_injection(self, format):
        """This test checks that we are able to read and fetch the injection bits
        from GWOSC files (see #1811).

        To do so, we download the injection data from GWOSC and compare it to a
        known StateVector.
        """
        inj = self.TEST_CLASS.fetch_open_data(
            GWOSC_GW150914_IFO,
            GWOSC_GW150914_SEGMENT[0],
            GWOSC_GW150914_SEGMENT[1],
            format=format,
            version=3,
            **GWOSC_GW150914_INJ_KWARGS[format],
        )
        # We know that the first bytes are equal to 23
        ref = StateVector(
            [23, 23, 23, 23],
            unit="",
            t0=GWOSC_GW150914_SEGMENT[0],
            dt=1,
            name=GWOSC_GW150914_INJ_NAME[format],
            bits=GWOSC_GW150914_INJ_BITS[format],
        )
        utils.assert_quantity_sub_equal(inj, ref, exclude=["channel", "bits"])
        assert sorted(map(str, inj.bits)) == sorted(map(str, ref.bits))

    @pytest.mark.parametrize("format", [
        "hdf5",
        pytest.param(  # only frameCPP actually reads units properly
            "gwf",
            marks=pytest.mark.requires("LDAStools.frameCPP"),
        ),
    ])
    @pytest_skip_flaky_network
    def test_fetch_open_data_injection_new_format(self, format):
        """Check that we can read and fetch GWOSC injection bits in the new format.

        To do so, we download the injection data from GWOSC and compare it to a
        known StateVector.

        See (https://gitlab.com/gwpy/gwpy/-/issues/1811).
        """
        inj = self.TEST_CLASS.fetch_open_data(
            GWOSC_GW190814_IFO,
            GWOSC_GW190814_SEGMENT[0],
            GWOSC_GW190814_SEGMENT[1],
            format=format,
            version=3,
            **GWOSC_GW190814_INJ_KWARGS[format],
        )
        # We know that the first bytes are equal to 23
        ref = StateVector(
            [23, 23, 23, 23],
            unit="",
            t0=GWOSC_GW190814_SEGMENT[0],
            dt=1,
            name=GWOSC_GW190814_INJ_NAME[format],
            bits=GWOSC_GW190814_INJ_BITS[format],
        )
        utils.assert_quantity_sub_equal(inj, ref, exclude=["channel", "bits"])
        assert sorted(map(str, inj.bits)) == sorted(map(str, ref.bits))


    @pytest.mark.requires("nds2")
    def test_from_nds2_buffer(self):
        # build fake buffer
        nds_buffer = mocks.nds2_buffer(
            "X1:TEST",
            self.data,
            1000000000,
            self.data.shape[0],
            "",
            name="test",
        )
        array = self.TEST_CLASS.from_nds2_buffer(nds_buffer)
        assert array.unit is units.dimensionless_unscaled

    @pytest.mark.parametrize("ext", ("hdf5", "h5"))
    def test_read_write_hdf5(self, tmp_path, array, ext):
        array.bits = ["a", "b", "c", "d"]
        array.name = "test"

        tmp = tmp_path / f"test.{ext}"

        # write array (with auto-identify)
        array.write(tmp, overwrite=True)

        # check reading gives the same data (with/without auto-identify)
        new = type(array).read(tmp, format="hdf5")
        utils.assert_quantity_sub_equal(array, new)
        new = type(array).read(tmp)
        utils.assert_quantity_sub_equal(array, new)


# -- StateVectorDict ----------------------------------------------------------

class TestStateVectorDict(_TestTimeSeriesBaseDict):
    TEST_CLASS = StateVectorDict
    ENTRY_CLASS = StateVector
    DTYPE = "uint32"


# -- StateVectorList ----------------------------------------------------------

class TestStateVectorList(_TestTimeSeriesBaseList):
    TEST_CLASS = StateVectorList
    ENTRY_CLASS = StateVector
    DTYPE = "uint32"



# Copyright (c) 2015-2017 Louisiana State University
#               2017-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""I/O routines for parsing Omega pipeline scan channel lists."""

from __future__ import annotations

import sys
from typing import TYPE_CHECKING

from ...io.utils import with_open
from .. import (
    Channel,
    ChannelList,
)

if TYPE_CHECKING:
    from typing import IO

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"


# -- read ----------------------------

@with_open
def read_omega_scan_config(source: IO) -> ChannelList:
    """Parse an Omega-scan configuration file into a `ChannelList`.

    Parameters
    ----------
    source : `str`
        path of Omega configuration file to parse

    Returns
    -------
    channels : `ChannelList`
        the list of channels (in order) as parsed

    Raises
    ------
    RuntimeError
        if this method finds a line it cannot parse sensibly
    """
    out = ChannelList()
    append = out.append
    section = None
    for line in source:
        line = line.strip()
        if isinstance(line, bytes):
            line = line.decode("utf-8")
        if not line or line.startswith("#"):
            continue
        if line.startswith("["):
            section = line[1:-1]
        elif line.startswith("{"):
            append(parse_omega_channel(source, section))
        else:
            msg = f"Failed to parse Omega config line: '{line}'"
            raise RuntimeError(msg)
    return out


def parse_omega_channel(
    fobj: IO,
    section: str | None = None,
) -> Channel:
    """Parse a `Channel` from an Omega-scan configuration file.

    Parameters
    ----------
    fobj : `file`
        the open file-like object to parse
    section : `str`
        name of section in which this channel should be recorded

    Returns
    -------
    channel : `Channel`
        the channel as parsed from this `file`
    """
    params = {}
    while True:
        line = next(fobj).strip()
        if isinstance(line, bytes):
            line = line.decode("utf-8")
        if not line:  # empty
            continue
        if line == "}":  # end of section
            break
        # parse 'key: value' setting
        key, value = line.split(":", 1)
        params[key.strip()] = omega_param(value)

    # build channel with params
    out = Channel(
        params.get("channelName"),
        sample_rate=params.get("sampleFrequency"),
        frametype=params.get("frameType"),
        frequency_range=params.get("searchFrequencyRange"),
    )
    out.group = section
    out.params = params

    return out


def omega_param(val: str) -> str | float | tuple[float, ...]:
    """Parse a value from an Omega-scan configuration file.

    This method tries to parse matlab-syntax parameters into a `str`,
    `float`, or `tuple`
    """
    val = val.strip()
    if val.startswith(('"', "'")):
        return str(val[1:-1])
    if val.startswith("["):
        return tuple(map(float, val[1:-1].split()))
    return float(val)


# -- write ---------------------------

@with_open(mode="w", pos=1)
def write_omega_scan_config(
    channellist: ChannelList,
    fobj: IO,
    *,
    header: bool = True,
) -> None:
    """Write a `ChannelList` to an Omega-pipeline scan configuration file.

    This method is dumb and assumes the channels are sorted in the right
    order already
    """
    # print header
    if header:
        print("# Q Scan configuration file", file=fobj)
        print("# Generated with GWpy from a ChannelList", file=fobj)
    group = None
    for channel in channellist:
        # print header
        if channel.group != group:
            group = channel.group
            print(f"\n[{group}]", file=fobj)
        print(file=fobj)
        print_omega_channel(channel, file=fobj)


def print_omega_channel(
    channel: Channel,
    file: IO = sys.stdout,
) -> None:
    """Print a `Channel` in Omega-pipeline scan format."""
    print("{", file=file)
    try:
        params = channel.params.copy()
    except AttributeError:
        params = {}
    params.setdefault("channelName", str(channel))
    params.setdefault("alwaysPlotFlag", int(params.pop("important", False)))
    if channel.frametype:
        params.setdefault("frameType", channel.frametype)
    if channel.sample_rate is not None:
        params.setdefault(
            "sampleFrequency",
            channel.sample_rate.to("Hz").value,
        )
    if channel.frequency_range is not None:
        low, high = channel.frequency_range.to("Hz").value
        params.setdefault("searchFrequencyRange", (low, high))
    if "qlow" in params or "qhigh" in params:
        qlow = params.pop("qlow", "sqrt(11)")
        qhigh = params.pop("qhigh", 64)
        params.setdefault("searchQRange", (qlow, qhigh))
    # write params
    for key in ["channelName", "frameType"]:
        if key not in params:
            msg = f"No '{key}' defined for {channel}"
            raise KeyError(msg)
    for key, value in params.items():
        keystr = f"{key}:"
        if isinstance(value, tuple):
            valuestr = f"[{' '.join(map(str, value))}]"
        elif isinstance(value, float) and value.is_integer():
            valuestr = str(int(value))
        elif isinstance(value, str):
            valuestr = repr(value)
        else:
            valuestr = str(value)
        print(f"  {keystr: <30}  {valuestr}", file=file)
    print("}", file=file)


# -- registry ------------------------

ChannelList.read.registry.register_reader(
    "omega-scan",
    ChannelList,
    read_omega_scan_config,
)
ChannelList.write.registry.register_writer(
    "omega-scan",
    ChannelList,
    write_omega_scan_config,
)



# Copyright (c) 2014-2017 Louisiana State University
#               2017-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""
.. sectionauthor:: Duncan Macleod <duncan.macleod@ligo.org>
.. currentmodule:: gwpy.table

Plotting an `EventTable` in a histogram
#######################################

I would like to study the distribution of the GW events detected to date.
"""

# %%
# First, we can download the ``'GWTC-1-confident'`` catalogue using
# :meth:`EventTable.fetch_open_data`:

from gwpy.table import EventTable
events = EventTable.fetch_open_data(
    "GWTC",
    columns=("mass_1_source", "mass_2_source"),
)
events.add_column(
    events["mass_1_source"] + events["mass_2_source"],
    name="mtotal",
)

# %%
# and can generate a new `~gwpy.plot.Plot` using the
# :meth:`~EventTable.hist` method:

plot = events.hist("mtotal", bins=20, range=(0, 100), histtype="stepfilled")
ax = plot.gca()
ax.set_xlabel(r"Total mass [M$_{\odot}$]")
ax.set_ylabel("Number of events")
ax.set_title("GWTC events")
plot.show()



# Copyright (c) 2014-2017 Louisiana State University
#               2017-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Read data from gravitational-wave frame (GWF) files using |LDAStools.frameCPP|__."""

from __future__ import annotations

import re
from collections import defaultdict
from math import ceil
from typing import TYPE_CHECKING

import numpy

from ....io.gwf import framecpp as io_framecpp
from ....io.utils import file_list
from ....segments import (
    Segment,
    SegmentList,
)
from ....time import (
    LIGOTimeGPS,
    to_gps,
)
from ... import (
    TimeSeries,
    TimeSeriesBaseList,
)
from ...core import _dynamic_scaled
from .utils import _channel_dict_kwarg

if TYPE_CHECKING:
    from collections.abc import (
        Callable,
        Iterator,
        Mapping,
    )
    from pathlib import Path
    from typing import (
        IO,
        Concatenate,
        TypeAlias,
        TypeVar,
    )

    import LDAStools.frameCPP

    from ....time import SupportsToGps
    from ... import (
        TimeSeriesBase,
        TimeSeriesBaseDict,
    )

    _TimeSeriesType = TypeVar("_TimeSeriesType", bound=TimeSeriesBase)
    _FrDataType: TypeAlias = (
        LDAStools.frameCPP.FrAdcData
        | LDAStools.frameCPP.FrProcData
        | LDAStools.frameCPP.FrSimData
    )

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"

# error regexs
FRERR_NO_FRAME_AT_NUM = re.compile(
    r"\ARequest for frame (?P<frnum>\d+) exceeds the range of "
    r"0 through (?P<nframes>\d+)\Z",
)
FRERR_NO_CHANNEL_OF_TYPE = re.compile(
    r"\ANo Fr(Adc|Proc|Sim)Data structures with the name ",
)

FRDATA_TYPES: tuple[str, ...] = (
    "ADC",
    "Proc",
    "Sim",
)


class _Skip(ValueError):  # noqa: N818
    """Error denoting that the contents of a given structure aren't required."""



# -- utilities -----------------------

def get_toc_segments(toc: LDAStools.frameCPP.FrTOC) -> SegmentList:
    """Return the segments for each frame listed in the TOC."""
    out = SegmentList()
    for s, ns, dur in zip(toc.GTimeS, toc.GTimeN, toc.dt, strict=True):
        fstart = LIGOTimeGPS(s, ns)
        out.append(Segment(fstart, fstart + dur))
    return out


def get_frame_segment(frame: LDAStools.frameCPP.FrameH) -> Segment:
    """Return the ``[start, end)`` segment for this frame."""
    epoch = LIGOTimeGPS(*frame.GetGTime())
    return Segment(epoch, epoch + frame.GetDt())


# -- read ----------------------------

def read(
    source: str | Path | IO,
    channels: list[str],
    start: SupportsToGps | None = None,
    end: SupportsToGps | None = None,
    scaled: bool | None = None,
    type: str | dict[str, str] | None = None,
    series_class: type[TimeSeriesBase] = TimeSeries,
) -> TimeSeriesBaseDict:
    """Read a dict of series from one or more GWF files.

    Parameters
    ----------
    source : `str`, `pathlib.path`, `file, `list`
        Source of data, any of the following:

        - `str` path of single data file,
        - `str` path of cache file,
        - `list` of paths.

    channels : `~gwpy.detector.ChannelList`, `list`
        A list of channels to read from the source.

    start : `~gwpy.time.LIGOTimeGPS`, `float`, `str` optional
        GPS start time of required data, anything parseable by
        :func:`~gwpy.time.to_gps` is fine.

    end : `~gwpy.time.LIGOTimeGPS`, `float`, `str`, optional
        GPS end time of required data, anything parseable by
        :func:`~gwpy.time.to_gps` is fine.

    scaled : `bool`, optional
        Apply slope and bias calibration to ADC data.

    type : `dict`, optional
        A `dict` of ``(name, channel-type)`` pairs, where ``channel-type``
        can be one of ``'adc'``, ``'proc'``, or ``'sim'``.

    series_class : `type`, optional
        The `Series` sub-type to return.

    Returns
    -------
    data : `~gwpy.timeseries.TimeSeriesDict` or similar
        A dict of ``(channel, series)`` pairs read from the GWF source(s).
    """
    # parse input source
    files = file_list(source)

    # parse type
    ctype = _channel_dict_kwarg(
        type,
        channels,
        expected_type=str,
        varname="type",
    )

    # read each file individually
    tmp: dict[str, TimeSeriesBaseList] = defaultdict(TimeSeriesBaseList)
    for file_ in files:
        # read each channel and each frame separately
        for series in _read_gwf(
            file_,
            channels,
            start,
            end,
            ctype,
            scaled,
            series_class,
        ):
            if series.name in tmp:
                tmp[series.name].append(series)
            else:
                tmp[series.name].append(numpy.require(series, requirements=["O"]))
    out = series_class.DictClass()
    for channel, serieslist in tmp.items():
        out[channel] = serieslist.join()

    return out


def _read_gwf(
    filename: str,
    channels: list[str],
    start: SupportsToGps | None,
    end: SupportsToGps | None,
    ctype: Mapping[str, str | None],
    scaled: bool | None,
    series_class: type[_TimeSeriesType],
) -> Iterator[_TimeSeriesType]:
    """Read a dict of series data from a single GWF file.

    Parameters
    ----------
    filename : `str`
        the GWF path from which to read

    channels : `~gwpy.detector.ChannelList`, `list`
        a list of channels to read from the source.

    start : `~gwpy.time.LIGOTimeGPS`, `float`, `str` optional
        GPS start time of required data, anything parseable by
        :func:`~gwpy.time.to_gps` is fine.

    end : `~gwpy.time.LIGOTimeGPS`, `float`, `str`, optional
        GPS end time of required data, anything parseable by
        :func:`~gwpy.time.to_gps` is fine.

    scaled : `bool`, optional
        apply slope and bias calibration to ADC data.

    ctype : `dict`, optional
        a `dict` of ``(name, channel-type)`` pairs, where ``channel-type``
        can be one of ``'adc'``, ``'proc'``, or ``'sim'``.

    series_class : `type`, optional
        the `Series` sub-type to return.

    Returns
    -------
    data : `~gwpy.timeseries.TimeSeriesDict` or similar
        a dict of ``(channel, series)`` pairs read from the GWF file.
    """
    # open file
    stream = io_framecpp.open_gwf(filename, "r")

    # parse start and end of data
    framesegments = get_toc_segments(stream.GetTOC())
    extent = framesegments.extent()
    if start is None:
        start = extent[0]
    else:
        start = to_gps(start)
    if end is None:
        end = extent[1]
    else:
        end = to_gps(end)

    # request segment
    span = Segment(start, end)

    # loop over frames in GWF
    record = set()
    for i, frameseg in enumerate(framesegments):
        # check whether we need this frame at all
        if not frameseg.intersects(span):
            continue

        # get data from each FrVect for each channel
        data = _read_frame(
            stream,
            i,
            frameseg[0],
            channels,
            start,
            end,
            scaled,
            ctype,
            series_class,
        )
        for series in data:
            yield series
            record.add(series.name)

    # if any channels weren't read, something went wrong
    for channel in channels:
        if (name := str(channel)) not in record:
            msg = f"failed to read '{name}' from '{filename}' in interval {span}"
            raise ValueError(msg)


def _read_frame(
    stream: LDAStools.frameCPP.IFrameFStream,
    index: int,
    epoch: LIGOTimeGPS,
    channels: list[str],
    start: LIGOTimeGPS,
    end: LIGOTimeGPS,
    scaled: bool | None,
    ctype: Mapping[str, str | None],
    series_class: type[_TimeSeriesType],
) -> Iterator[_TimeSeriesType]:
    """Read data from a specific frame in this stream."""
    # and read all the channels
    for channel in channels:
        _scaled = _dynamic_scaled(scaled, channel)
        try:
            new = _read_channel(
                stream,
                index,
                channel,
                ctype,
                epoch,
                start,
                end,
                scaled=_scaled,
                series_class=series_class,
            )
        except _Skip:  # don't need this frame for this channel
            continue
        yield from new


def _read_channel(
    stream: LDAStools.frameCPP.IFrameFStream,
    index: int,
    name: str,
    ctype: Mapping[str, str | None],
    epoch: LIGOTimeGPS,
    start: LIGOTimeGPS,
    end: LIGOTimeGPS,
    *,
    scaled: bool = False,
    series_class: type[_TimeSeriesType] = TimeSeries,
) -> Iterator[_TimeSeriesType]:
    """Read a channel from a specific frame in a stream."""
    data = _get_frdata(stream, index, name, ctype=ctype)
    return _read_frdata(
        data,
        epoch,
        start,
        end,
        scaled=scaled,
        series_class=series_class,
    )


def _get_frdata(
    stream: LDAStools.frameCPP.IFrameFStream,
    index: int,
    name: str,
    ctype: Mapping[str, str | None],
) -> _FrDataType:
    """Brute force-ish method to return the FrData structure for a channel.

    This saves on pulling the channel type from the TOC.
    """
    # types to try
    ctypes: tuple[str, ...]
    if ctp := ctype.get(name, None):
        ctypes = (ctp,)
    else:
        ctypes = FRDATA_TYPES

    # try each one
    for ctp in ctypes:
        try:
            _reader = getattr(stream, f"ReadFr{ctp.title()}Data")
        except AttributeError as exc:
            msg = f"invalid channel type '{ctp}' for {name}"
            raise ValueError(msg) from exc
        try:
            return _reader(index, name)
        except IndexError as exc:
            if FRERR_NO_CHANNEL_OF_TYPE.match(str(exc)):
                continue
            raise
    msg = f"channel '{name}' not found"
    raise ValueError(msg)


def _read_frdata(
    frdata: _FrDataType,
    epoch: LIGOTimeGPS,
    start: LIGOTimeGPS,
    end: LIGOTimeGPS,
    *,
    scaled: bool = False,
    series_class: type[_TimeSeriesType] = TimeSeries,
) -> Iterator[_TimeSeriesType]:
    """Read a series from an `FrData` structure.

    Parameters
    ----------
    frdata : `LDAStools.frameCPP.FrAdcData` or similar
        the data structure to read

    epoch : `float`
        the GPS start time of the containing frame
        (`LDAStools.frameCPP.FrameH.GTime`)

    start : `float`
        the GPS start time of the user request

    end : `float`
        the GPS end time of the user request

    scaled : `bool`, optional
        apply slope and bias calibration to ADC data.

    series_class : `type`, optional
        the `Series` sub-type to return.

    Returns
    -------
    series : `~gwpy.timeseries.TimeSeriesBase`
        the formatted data series

    Raises
    ------
    _Skip
        if this data structure doesn't overlap with the requested
        ``[start, end)`` interval.
    """
    datastart = epoch + frdata.GetTimeOffset()

    # check overlap with user-requested span
    if end and datastart >= end:
        raise _Skip

    # get scaling
    try:
        slope = frdata.GetSlope()
        bias = frdata.GetBias()
    except AttributeError:  # not FrAdcData
        slope = None
        bias = None
    else:
        # workaround https://git.ligo.org/ldastools/LDAS_Tools/-/issues/114
        # by forcing the default slope to 1.
        if bias == slope == 0.:
            slope = 1.
        null_scaling = slope == 1. and bias == 0.

    for j in range(frdata.data.size()):
        # we use range(frdata.data.size()) to avoid segfault
        # related to iterating directly over frdata.data
        try:
            new = _read_frvect(
                frdata.data[j],
                datastart,
                start,
                end,
                name=frdata.GetName(),
                series_class=series_class,
            )
        except _Skip:
            continue

        # apply scaling for ADC channels
        if scaled and slope is not None:
            rtype = numpy.result_type(new, slope, bias)
            typechange = not numpy.can_cast(
                rtype,
                new.dtype,
                casting="same_kind",
            )
            # only apply scaling if interesting _or_ if it would lead to a
            # type change, otherwise we are unnecessarily duplicating memory
            if typechange:
                new = new * slope + bias
            elif not null_scaling:
                new *= slope
                new += bias
        elif slope is not None:
            # user has deliberately disabled the ADC calibration, so
            # the stored engineering unit is not valid, revert to 'counts':
            new.override_unit("count")

        yield new


def _read_frvect(
    vect: LDAStools.frameCPP.FrVect,
    epoch: LIGOTimeGPS,
    start: LIGOTimeGPS,
    end: LIGOTimeGPS,
    name: str,
    series_class: type[_TimeSeriesType],
) -> _TimeSeriesType:
    """Read an array from an `FrVect` structure.

    Raises
    ------
    _Skip
        if this vect doesn't overlap with the requested
        ``[start, end)`` interval, or the name doesn't match.
    """
    # only read FrVect with matching name (or no name set)
    #    frame spec allows for arbitrary other FrVects
    #    to hold other information
    if vect.GetName() and name and vect.GetName() != name:
        raise _Skip

    # get array
    arr = vect.GetDataArray()
    nsamp = arr.size

    # and dimensions
    dim = vect.GetDim(0)
    dx = dim.dx
    x0 = dim.startX
    xunit = dim.GetUnitX() or None

    # start and end GPS times of this FrVect
    dimstart = epoch + x0
    dimend = dimstart + nsamp * dx

    # index of first required sample
    nxstart = int(max(0., float(start-dimstart)) / dx)

    # requested start time is after this frame, skip
    if nxstart >= nsamp:
        raise _Skip

    # index of end sample
    if end:
        nxend = int(nsamp - ceil(max(0., float(dimend-end)) / dx))
    else:
        nxend = None

    if nxstart or nxend:
        arr = arr[nxstart:nxend]

    # -- cast as a series

    # get unit
    unit = vect.GetUnitY() or None

    # create array
    series = series_class(
        numpy.require(arr, requirements=["O"]),
        t0=dimstart+nxstart*dx,
        dt=dx,
        name=name,
        channel=name,
        xunit=xunit,
        unit=unit,
        copy=False,
    )

    # add information to channel
    series.channel.sample_rate = series.sample_rate.value
    series.channel.unit = unit
    series.channel.dtype = series.dtype

    return series


# -- write ---------------------------

def write(
    tsdict: TimeSeriesBaseDict,
    outfile: str | Path | IO,
    start: LIGOTimeGPS,
    end: LIGOTimeGPS,
    type: str | None = None,
    name: str | None = None,
    run: int = 0,
    compression: int | str | None = None,
    compression_level: int | None = None,
) -> None:
    """Write data to a GWF file using the frameCPP API."""
    duration = end - start
    ifos = {
        ts.channel.ifo for ts in tsdict.values() if (
            ts.channel
            and ts.channel.ifo
            and ts.channel.ifo in io_framecpp.DetectorLocation.__members__
        )
    }

    # create frame
    frame = io_framecpp.create_frame(
        time=start,
        duration=duration,
        name=name or "gwpy",
        run=run,
        ifos=ifos,
    )

    # append channels
    for i, key in enumerate(tsdict):
        ctype = (
            type
            or getattr(tsdict[key].channel, "_ctype", "proc").lower()
            or "proc"
        )
        if ctype == "adc":
            kw = {"channelid": i}
        else:
            kw = {}
        _append_to_frame(
            frame,
            tsdict[key].crop(start, end),
            ctype=ctype,
            **kw,
        )

    # write frame to file
    io_framecpp.write_frames(
        outfile,
        [frame],
        compression=compression,
        compression_level=compression_level,
    )


def _append_to_frame(
    frame: LDAStools.frameCPP.FrameH,
    timeseries: TimeSeriesBase,
    ctype: str = "proc",
    **kwargs,
) -> None:
    """Append data from a `TimeSeries` to a `~frameCPP.FrameH`.

    Parameters
    ----------
    frame : `~frameCPP.FrameH`
        Frame header to append to.

    timeseries : `TimeSeries`
        The timeseries to append.

    ctype : `str`
        The type of the channel, one of 'adc', 'proc', 'sim'.

    kwargs
        Other keyword arguments are passed to the relevant
        `create_xxx` function.

    See Also
    --------
    gwpy.io.gwf.create_fradcdata
    gwpy.io.gwf.create_frprocdata
    gwpy.io.gwf_create_frsimdata
        For details of the data structure creation, and associated available
        arguments.
    """
    epoch = LIGOTimeGPS(*frame.GetGTime())

    # create the data container
    ctype = ctype.lower()
    create: Callable[Concatenate[TimeSeriesBase, ...], _FrDataType]
    if ctype == "adc":
        create = io_framecpp.create_fradcdata
        append = frame.AppendFrAdcData
    elif ctype == "proc":
        create = io_framecpp.create_frprocdata
        append = frame.AppendFrProcData
    elif ctype == "sim":
        create = io_framecpp.create_frsimdata
        append = frame.AppendFrSimData
    else:
        msg = (
            f"Invalid channel type '{ctype}', please select one of "
            "'adc, 'proc', or 'sim'",
        )
        raise RuntimeError(msg)
    frdata = create(
        timeseries,
        frame_epoch=epoch,
        **kwargs,
    )

    # append an FrVect
    frdata.AppendData(io_framecpp.create_frvect(timeseries))
    append(frdata)



# Copyright (c) 2019 Pensylvania State University
#               2019-2022 California Institute of Technology
#               2024-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Read events from the GstLAL online GW search."""

from __future__ import annotations

import re
from pathlib import Path
from typing import TYPE_CHECKING

from astropy.table import join

from ...io.ligolw import is_ligolw
from .. import EventTable
from .ligolw import read_table
from .utils import (
    DYNAMIC_COLUMN_FUNC,
    DYNAMIC_COLUMN_INPUT,
    dynamic_columns,
)

if TYPE_CHECKING:
    from collections.abc import (
        Callable,
        Collection,
        Iterable,
    )
    from typing import (
        IO,
        Literal,
    )

    from astropy.table import (
        Column,
        Table,
    )
    from igwn_ligolw.ligolw import Document

    from ...io.utils import (
        FileLike,
        FileSystemPath,
    )


__author__ = "Derk Davis <derek.davis@ligo.org>"
__credits__ = "Patrick Godwin <patrick.godwin@ligo.org>"

GSTLAL_FORMAT = "ligolw.gstlal"
GSTLAL_SNGL_FORMAT = "ligolw.gstlal.sngl"
GSTLAL_COINC_FORMAT = "ligolw.gstlal.coinc"

GSTLAL_FILENAME = re.compile("([A-Z][0-9])+-LLOID-[0-9.]+-[0-9.]+.xml.gz")


# -- identify ------------------------

def identify_gstlal(
    origin: Literal["read", "write"],
    filepath: FileSystemPath | None,
    fileobj: FileLike | None,
    *args,  # noqa: ANN002
    **kwargs,
) -> bool:
    """Identify a GstLAL file as a ligolw file with the correct name."""
    return (
        is_ligolw(origin, filepath, fileobj, *args, **kwargs)
        and (
            filepath is not None
            and GSTLAL_FILENAME.match(Path(filepath).name) is not None
        )
    )


# -- read ----------------------------

# singles format
def read_gstlal_sngl(
    source: str | Path | IO | Document | list[str | Path | IO],
    columns: Collection[str] | None = None,
    tablename: str = "sngl_inspiral",
    **kwargs,
) -> Table:
    """Read a `sngl_inspiral` table from one or more GstLAL LIGO_LW XML files.

    source : `file`, `str`, `~igwn_ligolw.ligolw.Document`, `list`
        One or more open files or file paths, or a single LIGO_LW ``Document``.

    kwargs
        Other keyword arguments are passed to `Table.read(format="ligolw")`.

    See Also
    --------
    gwpy.io.ligolw.read_table
        For details of keyword arguments for the read operation.

    gwpy.table.io.ligolw.to_astropy_table
        for details of keyword arguments for the conversion operation.
    """
    from igwn_ligolw import lsctables

    # handle column selection
    read_cols, dynamic_cols = dynamic_columns(
        columns,
        lsctables.TableByName[tablename].validcolumns,
        GET_COLUMN_EXTRA,
    )

    # read the table
    events = read_table(
        source,
        tablename=tablename,
        columns=read_cols or None,
        **kwargs,
    )

    if dynamic_cols:
        # generate requested derived columns on-the-fly
        for col_name in dynamic_cols:
            col_data = GET_COLUMN[col_name](events)
            events.add_column(col_data, name=col_name)
        # remove columns that were only added to generate a derived column
        for col_name in (read_cols or set()) - dynamic_cols:
            events.remove_column(col_name)

    return events


# coinc format
def read_gstlal_coinc(
    source: str | Path | IO | Document | list[str | Path | IO],
    columns: Collection[str] | None = None,
    **kwargs,
) -> Table:
    """Read a `Table` containing coinc event info from one or more GstLAL XML files.

    Parameters
    ----------
    source : `file`, `str`, `~igwn_ligolw.ligolw.Document`, `list`
        One or more open files or file paths, or a single LIGO_LW ``Document``.

    columns : `list` of `str`, optional
        List of column names to read. Valid column names are those in
        `coinc_inspiral` and `coinc_event` tables.

    kwargs
        Other keyword arguments are passed to `Table.read(format="ligolw")`.

    See Also
    --------
    gwpy.io.ligolw.read_table
        For details of keyword arguments for the read operation.

    gwpy.table.io.ligolw.to_astropy_table
        for details of keyword arguments for the conversion operation.
    """
    from igwn_ligolw.lsctables import TableByName
    extra_cols = set()

    if columns:
        # if columns were given, separate them into coinc_inspiral and
        # coinc_event columns
        columns = set(columns)
        val_col_inspiral = set(TableByName["coinc_inspiral"].validcolumns)
        val_col_event = set(TableByName["coinc_event"].validcolumns)
        for name in columns:
            if name not in (valid := val_col_inspiral | val_col_event):
                msg = (
                    f"'{name}' is not a valid column name. "
                    f"Valid column names: '{', '.join(valid)}'"
                )
                raise ValueError(msg)
        if "coinc_event_id" not in columns:
            columns.add("coinc_event_id")
            extra_cols.add("coinc_event_id")
        inspiral_cols = columns & val_col_inspiral
        inspiral_cols.add("coinc_event_id")
        event_cols = columns & val_col_event
    else:
        inspiral_cols = None
        event_cols = None

    # read tables
    coinc_inspiral = read_table(
        source,
        tablename="coinc_inspiral",
        columns=inspiral_cols,
        **kwargs,
    )
    coinc_event = read_table(
        source,
        tablename="coinc_event",
        columns=event_cols,
        **kwargs,
    )

    # join events based on coincs
    events = join(
        coinc_inspiral,
        coinc_event,
        keys="coinc_event_id",
        metadata_conflicts="silent",
    )
    events.meta["tablename"] = "gstlal_coinc_inspiral"

    # remove extra columns that we had to read but the user didn't ask for
    for col_name in extra_cols:
        events.remove_column(col_name)

    return events


# combined format
def read_gstlal(
    source: str | Path | IO | Document | list[str | Path | IO],
    triggers: str = "sngl",
    **kwargs,
) -> Table:
    """Read a `Table` from one or more GstLAL LIGO_LW XML files.

    source : `file`, `str`, `~igwn_ligolw.ligolw.Document`, `list`
        One or more open files or file paths, or a single LIGO_LW ``Document``.

    triggers : `str`, optional
        One of

        "sngl"
            For single-detector trigger information.

        "coinc"
            For coincident trigger information.

    kwargs
        Other keyword arguments are passed to `Table.read(format="ligolw")`.

    See Also
    --------
    gwpy.table.io.gstlal.read_gstlal_sngl
    gwpy.table.io.gstlal.read_gstlal_coinc
    """
    if triggers == "sngl":
        return read_gstlal_sngl(source, **kwargs)
    if triggers == "coinc":
        return read_gstlal_coinc(source, **kwargs)
    msg = f"triggers must be 'sngl' or 'coinc', got '{triggers}'"
    raise ValueError(msg)


# registers for unified I/O
EventTable.read.registry.register_identifier(
    GSTLAL_FORMAT,
    EventTable,
    identify_gstlal,
)
for fmt, reader in (
    (GSTLAL_SNGL_FORMAT, read_gstlal_sngl),
    (GSTLAL_COINC_FORMAT, read_gstlal_coinc),
    (GSTLAL_FORMAT, read_gstlal),
):
    EventTable.read.registry.register_reader(fmt, EventTable, reader)

# -- processed columns ---------------
#
# Here we define methods required to build commonly desired columns that
# are just a combination of the basic columns.
#
# Each method should take in a `~gwpy.table.Table` and return a
# `numpy.ndarray` or `~gwpy.table.Column`.

GET_COLUMN: dict[str, Callable] = {}
GET_COLUMN_EXTRA: dict[str, Iterable[str]] = {}


def get_snr_chi(
    events: Table,
    snr_pow: float = 2.,
    chi_pow: float = 2.,
) -> Column:
    """Calculate the 'SNR chi' column for this GstLAL ligolw table group."""
    snr = events["snr"][:]
    chisq = events["chisq"][:]
    return snr**snr_pow / chisq**(chi_pow/2.)


GET_COLUMN["snr_chi"] = get_snr_chi
GET_COLUMN_EXTRA["snr_chi"] = {"snr", "chisq"}


def get_chi_snr(
    events: Table,
    snr_pow: float = 2.,
    chi_pow: float = 2.,
) -> Column:
    """Calculate the 'chi SNR' column for this GstLAL ligolw table group.

    This is the reciprocal of the 'SNR chi' column.
    """
    return 1. / get_snr_chi(events, snr_pow, chi_pow)


GET_COLUMN["chi_snr"] = get_chi_snr
GET_COLUMN_EXTRA["chi_snr"] = {"snr", "chisq"}

# use the generic mass functions
for _key in (
    "mchirp",
    "mtotal",
):
    GET_COLUMN[_key] = DYNAMIC_COLUMN_FUNC[_key]
    GET_COLUMN_EXTRA[_key] = DYNAMIC_COLUMN_INPUT[_key]



# Copyright (c) 2021-2024 Evan Goetz
#               2021-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Transfer function plots."""

from __future__ import annotations

import logging
from collections import OrderedDict
from typing import TYPE_CHECKING

import numpy as np
from astropy.time import Time

from ..plot.bode import BodePlot
from ..plot.tex import label_to_latex
from .cliproduct import FrequencyDomainProduct

if TYPE_CHECKING:
    from argparse import (
        ArgumentParser,
        Namespace,
        _ArgumentGroup,
    )
    from logging import Logger
    from typing import ClassVar

    from ..frequencyseries import FrequencySeries
    from ..plot import Axes
    from ..segments import Segment
    from ..timeseries import TimeSeries

__author__ = "Evan Goetz <evan.goetz@ligo.org>"

logger = logging.getLogger(__name__)


class TransferFunctionProduct(FrequencyDomainProduct):
    """Plot transfer function between two series."""

    action: ClassVar[str] = "transferfunction"

    MIN_DATASETS = 2

    def __init__(
        self,
        args: Namespace,
        logger: Logger = logger,
    ) -> None:
        """Create a new `TransferFunctionProduct`."""
        super().__init__(args, logger=logger)
        self.ref_chan = self.args.ref or self.chan_list[0]
        # deal with channel type appendages
        self.ref_chan = self.ref_chan.split(",", maxsplit=1)[0]
        self.plot_dB = self.args.plot_dB
        self.test_chan = self.chan_list[1]
        self.tfs: list[FrequencySeries] = []

    @classmethod
    def arg_channels(cls, parser: ArgumentParser) -> _ArgumentGroup:
        """Add an `~argparse._ArgumentGroup` for channel options."""
        group = super().arg_channels(parser)
        group.add_argument(
            "--ref",
            help="Reference channel against which others will be compared",
        )
        return group

    @classmethod
    def arg_yaxis(cls, parser: ArgumentParser) -> _ArgumentGroup:
        """Add an `~argparse._ArgumentGroup` for Y-axis options."""
        group = cls._arg_axis("ymag", parser, scale="log")
        group.add_argument(
            "--plot-dB",
            action="store_true",
            help="Plot transfer function in dB",
        )
        return cls._arg_axis("yphase", parser, scale="linear")

    def _finalize_arguments(self, args: Namespace) -> None:
        """Finalise command-line arguments for this plot product."""
        if args.ymagscale is None:
            args.ymagscale = "log"
        if args.plot_dB:
            args.ymagscale = "linear"

        super()._finalize_arguments(args)

    def get_title(self) -> str:
        """Generate the title for this plot."""
        gps = self.start_list[0]
        utc = Time(gps, format="gps", scale="utc").iso
        tstr = f"{utc} | {gps} ({self.duration})"

        fftstr = f"fftlength={self.args.secpfft}, overlap={self.args.overlap}"

        return f"{tstr}, {fftstr}"

    def get_ylabel(self) -> str:
        """Text for y-axis label."""
        ylabelstr = ""
        if self.ax is self.axes[0]:
            ylabelstr = "Magnitude"
            if self.plot_dB:
                ylabelstr += " [dB]"
        else:
            ylabelstr = "Phase [deg.]"

        return ylabelstr

    def get_suptitle(self) -> str:
        """Start of default super title, first channel is appended to it."""
        return f"Transfer function: {self.test_chan}/{self.ref_chan}"

    def _set_axis_properties(
        self,
        ax: Axes,
        axis: str,
    ) -> None:
        """Set properties for X/Y axis on a specific subplot."""
        # Remove xlabel from top subplot
        if ax is self.axes[0] and axis == "x":
            ax.set_xlabel("")
        super()._set_axis_properties(ax, axis)

    def _set_yaxis_properties(self, ax: Axes) -> None:
        """Set properties for Y-axis."""
        if ax is self.axes[0]:
            axis_name = "ymag"
        else:
            axis_name = "yphase"
        self.plot.sca(ax)
        self._set_axis_properties(ax, axis_name)

    def _scale_axes_from_data(self, ax: Axes) -> None:
        """Restrict data limits for Y-axis based on what you can see."""
        axes_type = "mag" if ax is self.axes[0] else "phase"

        # get tight limits for X-axis
        xmin = self.args.xmin
        xmax = self.args.xmax
        if xmin is None:
            xmin = min(tf.xspan[0] for tf in self.tfs)
            # this is then typically zero, so if the xscale is log or None
            # we'll need to set to be the next bin higher (one step of df)
            if (
                xmin == 0
                and (self.args.xscale == "log" or self.args.xscale is None)
            ):
                xmin = min(tf.df.value for tf in self.tfs)
        if xmax is None:
            xmax = max(tf.xspan[1] for tf in self.tfs)

        # autoscale view for Y-axis
        cropped = [tf.crop(xmin, xmax) for tf in self.tfs]
        ymin = None
        ymax = None
        for tf in cropped:
            if axes_type == "mag" and self.plot_dB:
                vals = 20 * np.log10(abs(tf.value))
            elif axes_type == "mag":
                vals = abs(tf.value)
            else:
                vals = np.angle(tf.value, deg=True)
            minval = float(min(vals))
            maxval = float(max(vals))
            if ymin is None or minval < ymin:
                ymin = minval
            if ymax is None or maxval > ymax:
                ymax = maxval

        yint = ax.yaxis.get_data_interval()
        if ymin is None:
            ymin = yint[0]
        if ymax is None:
            ymax = yint[1]
        ax.yaxis.set_data_interval(ymin, ymax, ignore=True)
        ax.autoscale_view(scalex=False)

    def set_plot_properties(self) -> None:
        """Finalize figure object and show() or save()."""
        self.set_axes_properties()
        self.set_title(self.args.title, self.axes[0])
        self.set_suptitle(self.args.suptitle)
        enable = not self.args.nogrid
        self.set_grid(ax=self.axes[0], enable=enable)
        self.set_grid(ax=self.axes[1], enable=enable)

    def make_plot(self) -> BodePlot:
        """Generate the transfer function plot from the time series."""
        args = self.args

        fftlength = float(args.secpfft)
        overlap = args.overlap
        self.logger.debug(
            "Calculating transfer function secpfft: %f, overlap: %f",
            fftlength,
            overlap,
        )
        if overlap is not None:
            overlap *= fftlength

        self.logger.debug("Reference channel: %s", self.ref_chan)

        # group data by segment
        groups: dict[Segment, dict[str, TimeSeries]] = {}
        for series in self.timeseries:
            seg = series.span
            name = str(series.name or series.channel or "")
            try:
                groups[seg][name] = series
            except KeyError:
                groups[seg] = OrderedDict()
                groups[seg][name] = series

        # -- plot

        plot = BodePlot(
            figsize=self.figsize,
            dpi=self.dpi,
            dB=self.plot_dB,
        )
        self.tfs = []

        # calculate transfer function
        for datadict in groups.values():
            refts = datadict.pop(self.ref_chan)
            for name, series in datadict.items():
                self.test_chan = name
                tf = series.transfer_function(
                    refts,
                    fftlength=fftlength,
                    overlap=overlap,
                    window=args.window,
                )

                label = name
                if len(self.start_list) > 1:
                    label += f", {series.t0.value}"
                if self.usetex:
                    label = label_to_latex(label)

                plot.add_frequencyseries(tf, dB=self.plot_dB, label=label)
                self.tfs.append(tf)

        if args.xscale == "log" and not args.xmin:
            args.xmin = 1 / fftlength

        return plot



# Copyright (c) 2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Utility module to manage SciTokens for authorised data access.

The functions in this module require the following optional extra
packages:

- |htgettoken|_
- :doc:`igwn-auth-utils <igwn-auth-utils:index>`
- :doc:`requests-scitokens <requests-scitokens:index>`
- :doc:`scitokens <scitokens:index>`
"""

from __future__ import annotations

import logging
import sys
import tempfile
from contextlib import nullcontext
from pathlib import Path
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from collections.abc import MutableMapping
    from contextlib import AbstractContextManager

    from scitokens import SciToken

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"

log = logging.getLogger(__name__)

__all__ = [
    "add_http_authorization_header",
    "find_token",
    "get_scitoken",
]


# -- find an existing token

def find_token(
    audience: str,
    scope: str,
    **kwargs,
) -> SciToken:
    """Find and load a `Scitoken` for the given ``audience`` and ``scope``.

    Parameters
    ----------
    audience : `str`
        The required value for the ``aud`` claim.

    scope : `str`
        The required value for the ``scope`` claim. Can be a space-separated
        list of multiple scopes, _all_ of which must be matched by the token.

    kwargs
        All other keyword arguments are passed to
        `igwn_auth_utils.find_scitoken`.

    Returns
    -------
    token : `scitokens.SciToken`
        The token, if found, that matches the required claims.

    Raises
    ------
    igwn_auth_utils.IgwnAuthError
        If no valid token can be found.

    See Also
    --------
    igwn_auth_utils.find_scitoken
        For details on acceptable keyword arguments, and how tokens are
        discovered.

    gwpy.io.scitoken.get_scitoken
        To acquire a new token from the token issuer.
    """
    from igwn_auth_utils import find_scitoken
    return find_scitoken(
        audience,
        scope,
        **kwargs,
    )


# -- get a token

def _format_argv(**kwargs) -> list[str]:
    """Format arguments for ``htgettoken``."""
    args = []
    for key, value in kwargs.items():
        arg = f"--{key}"
        if value is False:  # disabled
            continue
        if value in (True, None):
            args.append(arg)
        else:
            args.extend((arg, str(value)))
    return args


def get_scitoken(
    *args: str,
    minsecs: float = 60,
    quiet: bool = True,
    **kwargs,
) -> SciToken:
    """Get a new `SciToken` using |htgettoken|_.

    Parameters
    ----------
    args
        All positional arguments are passed as arguments to
        `htgettoken.main`.

    minsecs : `float`, optional
        The minimum remaining lifetime to reuse an existing bearer token.

    quiet : `bool`, optional
        If `True`, supress output from `htgettoken`.

    kwargs
        All ``key: value`` keyword arguments (including ``minsecs`` and
        ``quiet``) are passed as ``--key=value`` options to
        `htgettoken.main`. Keywords with the value `True` are passed simply
        as ``--key``, while those with the value `False` are omitted.

    Returns
    -------
    token : `scitokens.SciToken`
        The new scitoken acquired by `htgettoken`.

    See Also
    --------
    htgettoken
    """
    import htgettoken
    from scitokens import SciToken

    if not sys.stdout.isatty():
        # don't prompt if we can't get a response
        kwargs.setdefault("nooidc", True)

    # parse output file if given
    outfile = kwargs.pop("outfile", None)

    # if not given, use a temporary directory
    ctx: AbstractContextManager
    if outfile is None:
        ctx = tempfile.TemporaryDirectory()
        filename = "test.token"
    else:
        outfile = Path(outfile)
        filename = outfile.name
        ctx = nullcontext(str(outfile.parent))

    # get token in a temporary directory
    with ctx as tmpdir:
        path = Path(tmpdir) / filename
        kwargs.setdefault("outfile", path)
        argv = list(args) + _format_argv(
            minsecs=minsecs,
            quiet=quiet,
            **kwargs,
        )
        try:
            htgettoken.main(argv)
        except SystemExit as exc:  # bad args
            msg = "htgettoken failed, see full traceback for details"
            raise RuntimeError(msg) from exc
        return SciToken.deserialize(path.read_text().strip())


# -- HTTP token handling -------------

def add_http_authorization_header(
    headers: MutableMapping,
    *,
    token: SciToken | None = None,
    error: bool = True,
    url: str | None = None,
    audience : str | None = None,
    create: bool = True,
    **claims,
) -> None:
    """Try and generate an HTTP ``Authorization`` header for a SciToken.

    This function tries to find an existing token that matches the required
    claims, otherwise will attempt to create (acquire) a new token.

    That token is then serialised to create the argument for an
    ``Authorization`` header using the ``Bearer`` scheme.

    Parameters
    ----------
    headers : `dict`
        The `dict` in which to store the ``Authorization`` header.

    url : `str`, optional
        The URL for which to attempt to find an existing token (to
        match the ``audience`` claim).

    error : `bool`, optional
        If `True` raise exceptions if tokens cannot be found or generated.

    token : `scitokens.SciToken`, optional
        A deserialised token to use for the header.

    audience : `str`, optional
        The required value for the ``aud`` claim.

    create : `bool`, optional
        If `True`, attempt to create (acquire) a new token if an existing
        token is not found.

    claims
        Other keyword arguments are passed as claims to
        `requests_scitokens.HTTPSciTokenAuth.find_token` or
        `gwpy.io.scitokens.get_scitoken`.

    See Also
    --------
    requests_scitokens.HTTPSciTokenAuth.find_token
        For details on how tokens are discovered.

    gwpy.io.scitokens.get_scitoken
        For details on how tokens are created (acquired).
    """
    error_obj: Exception | None = None
    error_types = (
        ImportError,
        RuntimeError,
        ValueError,
    )

    try:
        from requests_scitokens import HTTPSciTokenAuth
    except ImportError as exc:
        if error:
            raise
        log.debug("Cannot add SciToken Authorization header: %s", str(exc))
        return

    # construct the auth handler
    auth = HTTPSciTokenAuth(token=token, audience=audience)

    # try and find a token
    if auth.token is None:
        claims.setdefault("scope", None)
        # try and discover a token
        try:
            auth.token = auth.find_token(
                url=url,
                find_func=find_token,
                **claims,
            )
        except error_types as exc:
            error_obj = exc

    # try and create a token
    if auth.token is None and create:
        # try and generate a new token
        try:
            auth.token = get_scitoken(nooidc=True, **claims)
        except error_types as exc:
            error_obj = exc

    # if we got a token, serialize it and return
    if auth.token:
        headers["Authorization"] = auth.generate_bearer_header()
        log.debug("Added Authorization header with SciToken for %s", url)
        return

    # handle failure
    errmsg = "failed to identify a valid SciToken"
    if not error:
        log.debug("%s: %s", errmsg, str(error_obj))
        return
    raise RuntimeError(errmsg) from error_obj



# Copyright (c) 2017 Louisiana State University
#               2017-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Utilities for signal-processing with windows."""

from __future__ import annotations

import warnings
from math import ceil
from typing import TYPE_CHECKING

import numpy
from scipy.signal import (
    get_window as _get_window,
    windows as scipy_windows,
)
from scipy.special import expit

try:
    from scipy.signal.windows._windows import _WIN_FUNCS
except ImportError:  # scipy < 1.17
    try:
        from scipy.signal.windows._windows import (
            _win_equiv as WINDOWS,  # noqa: N812
        )
    except ImportError:  # Cannot find window equivalences
        WINDOWS = {}
else:
    WINDOWS = {name: func for name, (func, _) in _WIN_FUNCS.items()}

if TYPE_CHECKING:
    from typing import TypeAlias

    from numpy.typing import NDArray
    from optype.numpy import Array1D

    WindowLike: TypeAlias = str | float | tuple | Array1D[numpy.float64]

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"
__all__ = [
    "canonical_name",
    "get_window",
    "planck",
    "recommended_overlap",
]


def get_window(
    window: WindowLike,
    Nx: int,  # noqa: N803
    *args,
    **kwargs,
) -> NDArray:
    """Return a window of a given length and type.

    This is a thin wrapper around `scipy.signal.get_window` that handles
    pre-computed window arrays.

    Parameters
    ----------
    window : `str`, `float`, `tuple`, `numpy.ndarray`
        The specification for the window.
        Anything accepted by `scipy.signal.get_window` or an
        array-like object that is already a window (for convenience).

    Nx : `int`
        The size of the window.
        If ``window`` is an array, this size will be checked against the
        size of the array.

    args, kwargs
        All other arguments are passed to `scipy.signal.get_window`.

    Returns
    -------
    window : `numpy.ndarray`
        A 1-d window array with size ``Nx``.

    Raises
    ------
    ValueError
        If an ``window`` is given an array that doesn't have shape
        matching ``(Nx,)``.

    See Also
    --------
    scipy.signal.get_window
        For details of available window types and valid arguments.
    """
    # 1. try something floaty
    try:
        return _get_window(
            float(window),  # type: ignore[arg-type]
            Nx,
            *args,
            **kwargs,
        )
    except (TypeError, ValueError):
        pass

    # 2. try a name or tuple of params
    if isinstance(window, str | tuple):
        return _get_window(window, Nx, *args, **kwargs)

    # 3. otherwise we were something array-like
    window = numpy.asarray(window)

    # sanity check
    if window.shape != (Nx,):
        msg = f"invalid window array shape {window.shape}, should be ({Nx},)"
        raise ValueError(msg)

    return window


def canonical_name(name: str) -> str:
    """Find the canonical name for the given window in scipy.signal.

    Parameters
    ----------
    name : `str`
        The name of the window you want.

    Returns
    -------
    realname : `str`
        The name of the window as implemented in `scipy.signal.window`.

    Raises
    ------
    ValueError
        If ``name`` cannot be resolved to a window function in `scipy.signal`.

    Examples
    --------
    >>> from gwpy.signal.window import canonical_name
    >>> canonical_name("hann")
    'hann'
    >>> canonical_name("ksr")
    'kaiser'
    """
    # Strip any _symmetric or _periodic suffixes
    if name.endswith("_symmetric"):
        name = name[:-10]
    elif name.endswith("_periodic"):
        name = name[:-9]

    # Use equivalence introduced in scipy 0.16.0
    try:
        return WINDOWS[name.lower()].__name__
    except KeyError:  # no match
        if hasattr(scipy_windows, name):
            return name
        msg = f"no window function in scipy.signal equivalent to '{name}'"
        raise ValueError(msg) from None


# -- recommended overlap -------------
# source: http://edoc.mpg.de/395068

ROV: dict[str, float] = {
    "boxcar": 0,
    "bartlett": .5,
    "barthann": .5,
    "blackmanharris": .661,
    "flattop": .8,
    "hann": .5,
    "hamming": .5,
    "nuttall": .656,
    "triang": .5,
}


def recommended_overlap(
    name: str,
    nfft: int | None = None,
) -> float | int:
    """Return the recommended fractional overlap for the given window.

    If ``nfft`` is given, the return is in samples.

    Parameters
    ----------
    name : `str`
        The name of the window you are using.

    nfft : `int`
        The length of the window.

    Returns
    -------
    rov : `float`, `int`
        The recommended overlap (ROV) for the given window, in samples if
        ``nfft`` is given (`int`), otherwise fractional (`float`).

    Examples
    --------
    >>> from gwpy.signal.window import recommended_overlap
    >>> recommended_overlap("hann")
    0.5
    >>> recommended_overlap("blackmanharris", nfft=128)
    85
    """
    try:
        name = canonical_name(name)
    except KeyError as exc:
        raise ValueError(str(exc)) from None
    try:
        rov = ROV[name]
    except KeyError:
        msg = f"no recommended overlap for '{name}' window"
        raise ValueError(msg) from None
    if nfft:
        return ceil(nfft * rov)
    return rov


# -- Planck taper window -------------
# source: https://arxiv.org/abs/1003.2939

def planck(
    size: int,
    nleft: int = 0,
    nright: int = 0,
) -> NDArray:
    """Return a Planck taper window.

    .. deprecated:: 4.0.0

        This function is deprecated and will be removed in a future release.
        Use a different window function (e.g. `~scipy.signal.windows.tukey`)
        or implement your own version of the Planck window if needed.

    Parameters
    ----------
    size : `int`
        Number of samples in the output window.

    nleft : `int`
        Number of samples to taper on the left, should be less than `size/2`.

    nright : `int`
        Number of samples to taper on the right, should be less than `size/2`.

    Returns
    -------
    w : `ndarray`
        The window, with the maximum value normalized to 1 and at least one
        end tapered smoothly to 0.

    Examples
    --------
    To taper 0.1 seconds on both ends of one second of data sampled at 2048 Hz:

    >>> from gwpy.signal.window import planck
    >>> w = planck(2048, nleft=205, nright=205)

    References
    ----------
    .. [1] McKechan, D.J.A., Robinson, C., and Sathyaprakash, B.S. (April
           2010). "A tapering window for time-domain templates and simulated
           signals in the detection of gravitational waves from coalescing
           compact binaries". Classical and Quantum Gravity 27 (8).
           :doi:`10.1088/0264-9381/27/8/084020`

    .. [2] Wikipedia, "Window function",
           https://en.wikipedia.org/wiki/Window_function#Planck-taper_window
    """
    warnings.warn(
        "The planck window is deprecated and will be removed in a future "
        "release; consider implementing your own version if needed.",
        DeprecationWarning,
        stacklevel=2,
    )
    # construct a Planck taper window
    w = numpy.ones(size)
    if nleft:
        w[0] *= 0
        zleft = numpy.array([
            nleft * (1. / k + 1. / (k - nleft)) for k in range(1, nleft)
        ])
        w[1:nleft] *= expit(-zleft)
    if nright:
        w[size - 1] *= 0
        zright = numpy.array([
            -nright * (1. / (k - nright) + 1. / k) for k in range(1, nright)
        ])
        w[size - nright:size - 1] *= expit(-zright)
    return w

WINDOWS["planck"] = (planck, "OPTIONAL")



# Copyright (c) 2014-2017 Louisiana State University
#               2017-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Methods to calculate the sensitive distance."""

from __future__ import annotations

import warnings
from functools import wraps
from math import pi
from typing import (
    TYPE_CHECKING,
    cast,
)

from astropy import (
    constants,
    units,
)
from scipy.integrate import trapezoid
from scipy.interpolate import interp1d

from ..spectrogram import Spectrogram
from ..timeseries import TimeSeries
from ..utils import round_to_power

if TYPE_CHECKING:
    from collections.abc import Callable
    from typing import (
        ParamSpec,
        TypeVar,
    )

    import numpy
    from inspiral_range.waveform import CBCWaveform

    from ..frequencyseries import FrequencySeries

    P = ParamSpec("P")
    R = TypeVar("R")

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"
__credits__ = "Alex Urban <alexander.urban@ligo.org>"

DEFAULT_FFT_METHOD = "median"

PSD_UNIT = 1 / units.Hz


def _get_isco_frequency(
    mass1: float,
    mass2: float,
) -> units.Quantity:
    """Determine the innermost stable circular orbit (ISCO) frequency.

    Parameters
    ----------
    mass1 : `float`
        The mass (in solar masses) of the first binary component.

    mass2 : `float`
        The mass (in solar masses) of the second binary component.

    Returns
    -------
    fisco : `~astropy.units.Quantity`
        Linear frequency (Hz) of the innermost stable circular orbit.
    """
    mtotal = units.Quantity(mass1 + mass2, "solMass").to("kg")
    return (
        constants.c ** 3
        / (constants.G * 6 ** 1.5 * pi * mtotal)
    ).to("Hz")


def _get_spectrogram(
    hoft: TimeSeries | Spectrogram,
    **kwargs,
) -> Spectrogram:
    """Check that the input is a spectrogram, or compute one if compatible.

    Parameters
    ----------
    hoft : `~gwpy.timeseries.TimeSeries` or `~gwpy.spectrogram.Spectrogram`
        Record of gravitational-wave strain output from a detector.

    **kwargs
        Additional keyword arguments to
        `~gwpy.timeseries.TimeSeries.spectrogram`.

    Returns
    -------
    hoft : `~gwpy.spectrogram.Spectrogram`
        A time-frequency `Spectrogram` of the input.
    """
    if not isinstance(hoft, Spectrogram):
        try:
            hoft = hoft.spectrogram(**kwargs)
        except (
            AttributeError,  # object doesn't have a `.spectrogram()` method
            TypeError,  # something else went wrong
        ) as exc:
            msg = (
                "Could not produce a spectrogram from the input, please "
                "pass an instance of gwpy.timeseries.TimeSeries or "
                "gwpy.spectrogram.Spectrogram"
            )
            raise TypeError(msg) from exc
    return hoft


def _preformat_psd(
    func: Callable[P, R],
) -> Callable[P, R]:
    """Wrap a function to ensure that the incoming PSD has the right units."""

    @wraps(func)
    def decorated_func(*args: P.args, **kwargs: P.kwargs) -> R:
        psd = cast("FrequencySeries", args[0])
        # Force PSD to have the right units
        if psd.unit != PSD_UNIT:
            psd = psd.view()
            psd.override_unit(PSD_UNIT)
        args = (psd, *args[1:])  # type: ignore[assignment]
        return func(*args, **kwargs)
    return decorated_func


# -- SenseMon range ------------------

@_preformat_psd
def sensemon_range_psd(
    psd: FrequencySeries,
    snr: float = 8,
    mass1: float = 1.4,
    mass2: float = 1.4,
    *,
    horizon: bool = False,
) -> FrequencySeries:
    """Approximate the inspiral sensitive distance PSD from a GW strain PSD.

    This method returns the power spectral density (in ``Mpc**2 / Hz``) to
    which a compact binary inspiral with the given component masses would
    be detectable given the instrumental PSD. The calculation is defined in:
    :dcc:`LIGO-T030276`.

    Parameters
    ----------
    psd : `~gwpy.frequencyseries.FrequencySeries`
        The instrumental power-spectral-density data.

    snr : `float`, optional
        The signal-to-noise ratio for which to calculate range.

    mass1 : `float`, `~astropy.units.Quantity`, optional
        The mass of the first binary component
        (`float` assumed in solar masses).

    mass2 : `float`, `~astropy.units.Quantity`, optional
        The mass of the second binary component
        (`float` assumed in solar masses).

    horizon : `bool`, optional
        If `True`, return the maximal 'horizon' sensitive distance, otherwise
        return the angle-averaged range.

    Returns
    -------
    rspec : `~gwpy.frequencyseries.FrequencySeries`
        The calculated inspiral sensitivity PSD [Mpc^2 / Hz].
    """
    frange = (psd.frequencies > 0)
    # compute total mass and chirp mass
    mass1 = units.Quantity(mass1, "solMass").to("kg")
    mass2 = units.Quantity(mass2, "solMass").to("kg")
    mtotal = mass1 + mass2
    mchirp = (mass1 * mass2) ** (3 / 5.) / mtotal ** (1 / 5.)
    # calculate integrand with pre-factor
    prefactor = (
        (  # numerator
            (16 if horizon else 1.77 ** 2)
            * 5
            * constants.c ** (1 / 3.)
            * (mchirp * constants.G / constants.c ** 2) ** (5 / 3.)
        )  # denominator
        / (
            96
            * pi ** (4 / 3.)
            * snr ** 2
        )
    )
    return (  # inspiral range PSD, avoiding DC value
        1
        / psd[frange]
        * psd.frequencies[frange] ** (-7 / 3.)
        * prefactor
    ).to("Mpc^2 / Hz")


def sensemon_range(
    psd: FrequencySeries,
    snr: float = 8,
    mass1: float = 1.4,
    mass2: float = 1.4,
    fmin: float | None = None,
    fmax: float | None = None,
    *,
    horizon: bool = False,
) -> units.Quantity:
    """Approximate the inspiral sensitive distance from a GW strain PSD.

    This method returns the distance (in megaparsecs) to which a compact
    binary inspiral with the given component masses would be detectable
    given the instrumental PSD. The calculation is as defined in:
    :dcc:`LIGO-T030276`.

    Parameters
    ----------
    psd : `~gwpy.frequencyseries.FrequencySeries`
        The instrumental power-spectral-density data

    snr : `float`, optional
        The signal-to-noise ratio for which to calculate range.

    mass1 : `float`, `~astropy.units.Quantity`, optional
        The mass of the first binary component
        (`float` assumed in solar masses).

    mass2 : `float`, `~astropy.units.Quantity`, optional
        The mass of the second binary component
        (`float` assumed in solar masses).

    fmin : `float`, optional
        The lower frequency cut-off of the integral, default: `psd.df`.

    fmax : `float`, optional
        The maximum frequency limit of the integral, defaults to
        innermost stable circular orbit (ISCO) frequency

    horizon : `bool`, optional
        If `True`, return the maximal 'horizon' sensitive distance, otherwise
        return the angle-averaged range.

    Returns
    -------
    range : `~astropy.units.Quantity`
        the calculated inspiral range [Mpc]

    Examples
    --------
    Grab some data for LIGO-Livingston around GW150914 and generate a PSD:

    >>> from gwpy.timeseries import TimeSeries
    >>> hoft = TimeSeries.fetch_open_data("H1", 1126259446, 1126259478)
    >>> hoff = hoft.psd(fftlength=4)

    Now we can calculate the :func:`sensemon_range`:

    >>> from gwpy.astro import sensemon_range
    >>> r = sensemon_range(hoff, fmin: float = 30)
    >>> print(r)
    70.4612102889 Mpc
    """
    fisco = _get_isco_frequency(mass1, mass2)
    # format frequency limits
    fmin = units.Quantity(fmin or psd.df, "Hz")  # avoid DC value
    fmax = units.Quantity(fmax or fisco, "Hz")
    if fmax > fisco:
        warnings.warn(
            f"Upper frequency bound greater than {mass1}-{mass2} ISCO "
            f"frequency of {fisco}, using ISCO",
            stacklevel=2,
        )
        fmax = fisco
    # integrate and return
    f = psd.frequencies.to("Hz")
    frange = (f >= fmin) & (f < fmax)
    integrand = sensemon_range_psd(
        psd[frange],
        snr=snr,
        mass1=mass1,
        mass2=mass2,
        horizon=horizon,
    )
    return (units.Quantity(
        trapezoid(integrand.value, f.value[frange]),
        unit=integrand.unit * units.Hertz,
    ) ** (1 / 2.)).to("Mpc")


# -- inspiral range ------------------

MISSING_INSPIRAL_RANGE_MESSAGE = (
    "gwpy.astro's inspiral_range and inspiral_range_psd functions "
    "require the extra package 'inspiral-range' to provide "
    "cosmologically-corrected distance calculations, please install "
    "that package and try again, or install gwpy with the 'astro' "
    "extra via `python -m pip install gwpy[astro]`"
)


def _cbc_waveform(
    frequencies: numpy.ndarray,
    m1: float,
    m2: float,
    **kwargs,
) -> CBCWaveform:
    """Generate a CBCWaveform."""
    try:
        from inspiral_range.waveform import CBCWaveform
    except ImportError as exc:
        exc.args = (f"{exc}; {MISSING_INSPIRAL_RANGE_MESSAGE}",)
        raise

    return CBCWaveform(
        frequencies,
        m1=m1,
        m2=m2,
        **kwargs,
    )


@_preformat_psd
def inspiral_range_psd(
    psd: FrequencySeries,
    snr: float = 8,
    mass1: float = 1.4,
    mass2: float = 1.4,
    *,
    horizon: bool = False,
    **kwargs,
) -> FrequencySeries:
    """Calculate the cosmology-corrected inspiral sensitive distance PSD.

    This method returns the power spectral density (in ``Mpc**2 / Hz``) to
    which a compact binary inspiral with the given component masses would
    be detectable given the instrumental PSD. The calculation is defined in
    Belczynski et. al (2014):

    https://dx.doi.org/10.1088/0004-637x/789/2/120

    Parameters
    ----------
    psd : `~gwpy.frequencyseries.FrequencySeries`
        The instrumental power-spectral-density data.

    snr : `float`, optional
        The signal-to-noise ratio for which to calculate range.

    mass1 : `float`, `~astropy.units.Quantity`, optional
        The mass of the first binary component
        (`float` assumed in solar masses).

    mass2 : `float`, `~astropy.units.Quantity`, optional
        The mass of the second binary component
        (`float` assumed in solar masses).

    horizon : `bool`, optional
        If `True`, return the maximal 'horizon' luminosity distance, otherwise
        return the angle-averaged comoving distance.

    **kwargs
        Additional keyword arguments to `~inspiral_range.waveform.CBCWaveform`.

    Returns
    -------
    rspec : `~gwpy.frequencyseries.FrequencySeries`
        The calculated inspiral sensitivity PSD [Mpc^2 / Hz].

    See Also
    --------
    sensemon_range_psd
        For the method based on :dcc:`LIGO-T030276`, also known as LIGO SenseMonitor.

    inspiral-range
        The package which does heavy lifting for waveform simulation and
        cosmology calculations.
    """
    try:
        from inspiral_range import (
            find_root_redshift,
            range as range_func,
        )
    except ImportError as exc:
        exc.msg = f"{exc}; {MISSING_INSPIRAL_RANGE_MESSAGE}"
        raise

    f = psd.frequencies.to("Hz").value
    frange = (f > 0)

    # generate a waveform for this system
    inspiral = _cbc_waveform(
        f[frange],
        m1=mass1,
        m2=mass2,
        **kwargs,
    )

    # determine the detector horizon redshift
    z_hor = find_root_redshift(
        lambda z: inspiral.SNR(psd.value[frange], z) - snr,
    )
    if horizon:
        dist = inspiral.cosmo.luminosity_distance(z_hor)
    else:
        dist = range_func(
            f[frange],
            psd.value[frange],
            z_hor=z_hor,
            H=inspiral,
            detection_snr=snr,
        )

    # calculate the sensitive distance PSD
    (fz, hz) = inspiral.z_scale(z_hor)
    hz = interp1d(
        fz,
        hz,
        bounds_error=False,
        fill_value=(hz[0], 0),
    )(f)
    out = type(psd)(
        4
        * (dist / snr)**2
        * (hz**2 / psd.value)[f > 0],
    )

    # finalize properties and return
    out.__array_finalize__(psd)
    out.override_unit("Mpc^2 / Hz")
    out.f0 = f[f > 0][0]
    return out


@_preformat_psd
def inspiral_range(
    psd: FrequencySeries,
    snr: float = 8,
    mass1: float = 1.4,
    mass2: float = 1.4,
    fmin: float | None = None,
    fmax: float | None = None,
    *,
    horizon: bool = False,
    **kwargs,
) -> units.Quantity:
    """Calculate the cosmology-corrected inspiral sensitive distance.

    This method returns the distance (in megaparsecs) to which a compact
    binary inspiral with the given component masses would be detectable
    given the instrumental PSD. The calculation is defined in Belczynski
    et. al (2014):

    https://dx.doi.org/10.1088/0004-637x/789/2/120

    Parameters
    ----------
    psd : `~gwpy.frequencyseries.FrequencySeries`
        The instrumental power-spectral-density data.

    snr : `float`, optional
        The signal-to-noise ratio for which to calculate range.

    mass1 : `float`, `~astropy.units.Quantity`, optional
        The mass of the first binary component
        (`float` assumed in solar masses).

    mass2 : `float`, `~astropy.units.Quantity`, optional
        The mass of the second binary component
        (`float` assumed in solar masses).

    fmin : `float`, optional
        The lower frequency cut-off of the integral, default: `psd.df`.

    fmax : `float`, optional
        The maximum frequency limit of the integral, defaults to the rest-frame
        innermost stable circular orbit (ISCO) frequency.

    horizon : `bool`, optional
        If `True`, return the maximal 'horizon' luminosity distance, otherwise
        return the angle-averaged comoving distance.

    **kwargs
        Additional keyword arguments to `~inspiral_range.waveform.CBCWaveform`.

    Returns
    -------
    range : `~astropy.units.Quantity`
        The calculated inspiral range [Mpc].

    Examples
    --------
    Grab some data for LIGO-Livingston around GW150914 and generate a PSD:

    >>> from gwpy.timeseries import TimeSeries
    >>> hoft = TimeSeries.fetch_open_data("H1", 1126259446, 1126259478)
    >>> hoff = hoft.psd(fftlength=4)

    Now, we can calculate the :func:`inspiral_range`:

    >>> from gwpy.astro import inspiral_range
    >>> r = inspiral_range(hoff, fmin: float = 30)
    >>> print(r)
    70.4612102889 Mpc

    See Also
    --------
    sensemon_range
        For the method based on :dcc:`LIGO-T030276`, also known as LIGO SenseMonitor.

    inspiral-range
        The package which does heavy lifting for waveform simulation and
        cosmology calculations.
    """
    try:
        from inspiral_range import (
            find_root_redshift,
            range as range_func,
        )
    except ImportError as exc:
        exc.msg = f"{exc}; {MISSING_INSPIRAL_RANGE_MESSAGE}"
        raise

    # format frequency limits
    f = psd.frequencies.to("Hz").value
    fmin = fmin or psd.df.value  # avoid DC value
    fmax = fmax or round_to_power(f[-1], which="lower")
    frange = (f >= fmin) & (f < fmax)

    # generate a waveform for this system
    inspiral = _cbc_waveform(
        f[frange],
        m1=mass1,
        m2=mass2,
        **kwargs,
    )

    # determine the detector horizon redshift
    z_hor = find_root_redshift(
        lambda z: inspiral.SNR(psd.value[frange], z=z) - snr,
    )

    # return the sensitive distance metric
    if horizon:
        dist = inspiral.cosmo.luminosity_distance(z_hor)
    else:
        dist = range_func(
            f[frange],
            psd.value[frange],
            z_hor=z_hor,
            H=inspiral,
            detection_snr=snr,
        )
    return units.Quantity(dist, unit="Mpc")


# -- burst range ---------------------

@_preformat_psd
def burst_range_spectrum(
    psd: FrequencySeries,
    snr: float = 8,
    energy: float = 1e-2,
) -> FrequencySeries:
    """Calculate the frequency-dependent GW burst range from a strain PSD.

    Parameters
    ----------
    psd : `~gwpy.frequencyseries.FrequencySeries`
        The instrumental power-spectral-density data.

    snr : `float`, optional
        The signal-to-noise ratio for which to calculate range.

    energy : `float`, optional
        The relative energy output of the GW burst.

    Returns
    -------
    rangespec : `~gwpy.frequencyseries.FrequencySeries`
        The burst range `FrequencySeries` [Mpc (default)].
    """
    frange = (psd.frequencies > 0)
    # calculate frequency dependent range in parsecs
    a = (
        constants.G
        * energy
        * constants.M_sun
        * 0.4
        / (pi**2 * constants.c)
    ) ** (1 / 2.)
    return (  # burst range spectrum, avoiding DC value
        psd[frange] ** (-1 / 2.)
        * a
        / (snr * psd.frequencies[frange])
    ).to("Mpc")


def burst_range(
    psd: FrequencySeries,
    snr: float = 8,
    energy: float = 1e-2,
    fmin: float = 100,
    fmax: float = 500,
) -> units.Quantity:
    """Calculate the integrated GRB-like GW burst range from a strain PSD.

    Parameters
    ----------
    psd : `~gwpy.frequencyseries.FrequencySeries`
        The instrumental power-spectral-density data.

    snr : `float`, optional
        The signal-to-noise ratio for which to calculate range.

    energy : `float`, optional
        The relative energy output of the GW burst.

    fmin : `float`, optional
        The lower frequency cutoff of the burst range integral.

    fmax : `float`, optional
        The upper frequency cutoff of the burst range integral.

    Returns
    -------
    range : `~astropy.units.Quantity`
        The GRB-like-burst sensitive range [Mpc (default)].

    Examples
    --------
    Grab some data for LIGO-Livingston around GW150914 and generate a PSD

    >>> from gwpy.timeseries import TimeSeries
    >>> hoft = TimeSeries.fetch_open_data("H1", 1126259446, 1126259478)
    >>> hoff = hoft.psd(fftlength=4)

    Now we can calculate the :func:`burst_range`:

    >>> from gwpy.astro import burst_range
    >>> r = burst_range(hoff, fmin: float = 30)
    >>> print(r)
    42.5055584195 Mpc
    """
    f = psd.frequencies.to("Hz").value
    # restrict integral
    fmin = fmin or psd.df.to("Hz").value
    fmax = fmax or f[-1].value
    frange = (f >= fmin) & (f < fmax)
    # calculate integrand and integrate
    integrand = burst_range_spectrum(
        psd[frange],
        snr=snr,
        energy=energy,
    ) ** 3
    out = trapezoid(integrand.value, f[frange])
    # normalize and return
    return (units.Quantity(
        out / (fmax - fmin),
        unit=integrand.unit,
    ) ** (1 / 3.)).to("Mpc")


# -- timeseries/spectrogram wrappers -

def range_timeseries(
    hoft: TimeSeries | Spectrogram,
    stride: float | None = None,
    fftlength: float | None = None,
    overlap: float | None = None,
    window: str | numpy.ndarray = "hann",
    method: str = DEFAULT_FFT_METHOD,
    nproc: int = 1,
    range_func: Callable | None = None,
    **rangekwargs,
) -> TimeSeries:
    """Estimate timeseries of astrophysical range (Mpc).

    Parameters
    ----------
    hoft : `~gwpy.timeseries.TimeSeries`, `~gwpy.spectrogram.Spectrogram`
        Detector (strain) data from which to estimate range.

    stride : `float`, optional
        Desired step size (seconds) of range timeseries, required if
        `hoft` is an instance of `TimeSeries`.

    fftlength : `float`, optional
        Number of seconds in a single FFT.

    overlap : `float`, optional
        Number of seconds of overlap between FFTs, defaults to the
        recommended overlap for the given window (if given), or ``0``.

    window : `str`, `numpy.ndarray`, optional
        Window function to apply to timeseries prior to FFT, see
        :func:`scipy.signal.get_window` for details on acceptable
        formats.

    method : `str`, optional
        FFT-averaging method, defaults to median averaging, see
        :meth:`~gwpy.timeseries.TimeSeries.spectrogram` for
        more details

    nproc : `int`, optional
        number of CPUs to use in parallel processing of FFTs, default: 1

    range_func : `callable`, optional
        the function to call to generate the range for each stride,
        defaults to ``inspiral_range`` unless ``energy`` is given
        as a keyword argument to the range function

    **rangekwargs : `dict`, optional
        additional keyword arguments to :func:`burst_range` or
        :func:`inspiral_range` (see "Notes" below), defaults to
        inspiral range with `mass1 = mass2 = 1.4` solar masses

    Returns
    -------
    out : `~gwpy.timeseries.TimeSeries`
        timeseries trends of astrophysical range

    Notes
    -----
    This method is designed to quantify a gravitational-wave detector's
    sensitive range as a function of time. It supports the range to
    compact binary inspirals and to unmodelled GW bursts, each a class
    of transient event.

    See Also
    --------
    gwpy.timeseries.TimeSeries.spectrogram
        for the underlying power spectral density estimator
    inspiral_range
        for the function that computes inspiral range
    burst_range
        for the function that computes burst range
    range_spectrogram
        for a `~gwpy.spectrogram.Spectrogram` of the range integrand
    """
    if not isinstance(hoft, Spectrogram):
        hoft = _get_spectrogram(
            hoft,
            stride=stride,
            fftlength=fftlength,
            overlap=overlap,
            window=window,
            method=method,
            nproc=nproc,
        )

    rangekwargs = rangekwargs or {
        "mass1": 1.4,
        "mass2": 1.4,
    }
    if range_func is None and "energy" in rangekwargs:
        range_func = burst_range
    elif range_func is None:
        range_func = inspiral_range

    # calculate the timeseries of range
    out = TimeSeries(
        [range_func(psd, **rangekwargs).value for psd in hoft],
    )
    out.__array_finalize__(hoft)
    out.override_unit("Mpc")
    return out


def range_spectrogram(
    hoft: TimeSeries | Spectrogram,
    stride: float | None = None,
    fftlength: float | None = None,
    overlap: float | None = None,
    window: str | numpy.ndarray = "hann",
    method: str = DEFAULT_FFT_METHOD,
    nproc: int = 1,
    range_func: Callable | None = None,
    **rangekwargs,
) -> Spectrogram:
    """Calculate the average range spectrogram (Mpc or Mpc^2 / Hz) from strain.

    Parameters
    ----------
    hoft : `~gwpy.timeseries.TimeSeries`  or `~gwpy.spectrogram.Spectrogram`
        record of gravitational-wave strain output from a detector

    stride : `float`, optional
        number of seconds in a single PSD (i.e., step size of spectrogram),
        required if `hoft` is an instance of `TimeSeries`

    fftlength : `float`, optional
        number of seconds in a single FFT

    overlap : `float`, optional
        number of seconds of overlap between FFTs, defaults to the
        recommended overlap for the given window (if given), or 0

    window : `str`, `numpy.ndarray`, optional
        window function to apply to timeseries prior to FFT, see
        :func:`scipy.signal.get_window` for details on acceptable
        formats

    method : `str`, optional
        FFT-averaging method, defaults to median averaging, see
        :meth:`~gwpy.timeseries.TimeSeries.spectrogram` for
        more details

    nproc : `int`, optional
        number of CPUs to use in parallel processing of FFTs, default: 1

    fmin : `float`, optional
        low frequency cut-off (Hz), defaults to `1/fftlength`

    fmax : `float`, optional
        high frequency cut-off (Hz), defaults to Nyquist frequency of `hoft`

    range_func : `callable`, optional
        the function to call to generate the range for each stride,
        defaults to ``inspiral_range`` unless ``energy`` is given
        as a keyword argument to the range function

    **rangekwargs : `dict`, optional
        additional keyword arguments to :func:`burst_range_spectrum` or
        :func:`inspiral_range_psd` (see "Notes" below), defaults to
        inspiral range with `mass1 = mass2 = 1.4` solar masses

    Returns
    -------
    out : `~gwpy.spectrogram.Spectrogram`
        time-frequency spectrogram of astrophysical range

    Notes
    -----
    This method is designed to show the contribution to a
    gravitational-wave detector's sensitive range across frequency bins
    as a function of time. It supports the range to compact binary
    inspirals and to unmodelled GW bursts, each a class of transient
    event.

    If inspiral range is requested and `fmax` exceeds the frequency of the
    innermost stable circular orbit (ISCO), the output will extend only up
    to the latter.

    See Also
    --------
    gwpy.timeseries.TimeSeries.spectrogram
        for the underlying power spectral density estimator
    inspiral_range_psd
        for the function that computes inspiral range integrand
    burst_range_spectrum
        for the function that computes burst range integrand
    range_timeseries
        for `TimeSeries` trends of the astrophysical range
    """
    rangekwargs = rangekwargs or {
        "mass1": 1.4,
        "mass2": 1.4,
    }
    if range_func is None and "energy" in rangekwargs:
        range_func = burst_range_spectrum
    elif range_func is None:
        range_func = inspiral_range_psd
    hoft = _get_spectrogram(
        hoft,
        stride=stride,
        fftlength=fftlength,
        overlap=overlap,
        window=window,
        method=method,
        nproc=nproc,
    )
    # set frequency limits
    f = hoft.frequencies.to("Hz")
    fmin = units.Quantity(
        rangekwargs.pop("fmin", hoft.df),
        "Hz",
    )
    fmax = units.Quantity(
        rangekwargs.pop(
            "fmax",
            round_to_power(f[-1].value, which="lower"),
        ),
        "Hz",
    )
    frange = (f >= fmin) & (f < fmax)
    # loop over time bins
    out = Spectrogram(
        [range_func(psd[frange], **rangekwargs).value for psd in hoft],
    )
    # finalise output
    out.__array_finalize__(hoft)
    out.override_unit("Mpc" if "energy" in rangekwargs else "Mpc^2 / Hz")
    out.f0 = fmin
    return out



# Copyright (c) 2014-2017 Louisiana State University
#               2017-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Unit test for utils module."""

import numpy
import pytest
from astropy import units

# import necessary modules
lal = pytest.importorskip("lal")
utils_lal = pytest.importorskip("gwpy.utils.lal")

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"


@pytest.mark.parametrize(("arg", "result"), [
    (float, "REAL8"),
    (numpy.dtype("float64"), "REAL8"),
    (11, "REAL8"),
])
def test_to_lal_type_str(arg, result):
    """Test :func:`gwpy.utils.lal.to_lal_type_str`."""
    assert utils_lal.to_lal_type_str(arg) == result


@pytest.mark.parametrize("arg", [
    "blah",
    numpy.int8,
    20,
])
def test_to_lal_type_str_error(arg):
    """Test :func:`gwpy.utils.lal.to_lal_type_str` error handling."""
    with pytest.raises(
        ValueError,
        match="Failed to map",
    ):
        utils_lal.to_lal_type_str(arg)


def test_find_typed_function():
    """Test :func:`gwpy.utils.lal.find_typed_function`."""
    assert utils_lal.find_typed_function(
        "REAL8",
        "Create",
        "Sequence",
    ) is lal.CreateREAL8Sequence


@pytest.mark.requires("lalframe")
def test_find_typed_function_module():
    """Test :func:`gwpy.utils.lal.find_typed_function` with lalframe."""
    import lalframe

    assert utils_lal.find_typed_function(
        "REAL4",
        "FrStreamRead",
        "TimeSeries",
        module=lalframe,
    ) is lalframe.FrStreamReadREAL4TimeSeries


@pytest.mark.parametrize("dtype", [
    numpy.int16,
    numpy.int32,
    numpy.int64,
    numpy.uint16,
    numpy.uint32,
    numpy.uint64,
    numpy.float32,
    numpy.float64,
    numpy.complex64,
    numpy.complex128,
])
def test_from_lal_type(dtype):
    """Test :func:`gwpy.utils.lal.from_lal_type`."""
    func = utils_lal.find_typed_function(
        dtype,
        "Create",
        "TimeSeries",
    )
    lalts = func(
        None,
        None,
        0,
        1.0,
        None,
        1,
    )
    assert utils_lal.from_lal_type(lalts) is dtype
    assert utils_lal.from_lal_type(type(lalts)) is dtype


@pytest.mark.parametrize("name", [
    "XXX",
    "XXXCOMPLEX64ZZZ",
    "INNT2",
    "INT5",
    "REAL42",
])
def test_from_lal_type_errors(name):
    """Test :func:`gwpy.utils.lal.from_lal_type` error handling."""
    lal_type = type(name, (), {})
    with pytest.raises(
        ValueError,
        match="no known numpy",
    ):
        utils_lal.from_lal_type(lal_type)


@pytest.mark.parametrize(("in_", "out", "scale"), [
    ("m", "m", 1),
    ("Farad", "m^-2 kg^-1 s^4 A^2", 1),
    ("m**(1/2)", "m^1/2", 1),
    ("km", "10^3 m", 1),
    ("123 m", "m", 123),
])
def test_to_lal_unit(in_, out, scale):
    """Test `to_lal_unit`."""
    assert utils_lal.to_lal_unit(in_) == (lal.Unit(out), scale)


def test_to_lal_unit_error():
    """Test `to_lal_unit` error handling."""
    with pytest.raises(
        ValueError,
        match=r"^LAL has no unit corresponding to 'rad'$",
    ):
        utils_lal.to_lal_unit("rad/s")


@pytest.mark.parametrize(("in_", "out"), [
    ("m s^-1", "m/s"),
    ("strain", "strain"),
    ("m^1/2", "m**(1/2.)"),
    ("10^3 m", "km"),
])
def test_from_lal_unit(in_, out):
    """Test :func:`gwpy.utils.lal.from_lal_unit`."""
    assert utils_lal.from_lal_unit(
        lal.Unit(in_),
    ) == units.Unit(out)


def test_to_lal_ligotimegps():
    """Test :func:`gwpy.utils.lal.to_lal_ligotimegps`."""
    assert utils_lal.to_lal_ligotimegps(
        123.456,
    ) == lal.LIGOTimeGPS(123, 456000000)



# Copyright (c) 2015-2020 Joseph Areeda
#               2020-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Generate plots of GW observatory data using GWpy."""

from __future__ import annotations

import os
import sys
import time
from typing import TYPE_CHECKING

from matplotlib import use

from .. import __version__
from ..tools import _utils
from . import PRODUCTS

if TYPE_CHECKING:
    from argparse import (
        ArgumentParser,
        Namespace,
    )

__author__ = "Joseph Areeda <joseph.areeda@ligo.org>"

logger = _utils.get_logger(__name__)

# if launched from a terminal with no display
# Must be done before modules like pyplot are imported
if len(os.getenv("DISPLAY", "")) == 0:
    use("Agg")

INTERACTIVE = hasattr(sys, "ps1")

EPILOG = f"""
Examples:

    $ gwpy-plot timeseries --chan H1:GDS-CALIB_STRAIN --start 1126259457

    $ gwpy-plot spectrum --chan H1:GDS-CALIB_STRAIN L1:GDS-CALIB_STRAIN --chan V1:Hrec_hoft_16384Hz --start 1187008866 --duration 32 --xmin 10 --xmax 4000

    $ gwpy-plot coherencegram --chan H1:GDS-CALIB_STRAIN H1:PEM-CS_ACC_PSL_PERISCOPE_X_DQ --start 1126260017 --duration 600

Written by {__author__}.
Report bugs to https://gitlab.com/gwpy/gwpy/-/issues/.
"""  # noqa: E501


# -- init command line ---------------

def create_parser() -> ArgumentParser:
    """Create the command line argument parser for `gwpy-plot`."""
    parser = _utils.ArgumentParser(
        description=__doc__,
        epilog=EPILOG,
    )
    parser.add_argument(
        "-V", "--version",
        action="version",
        version=__version__,
        help="show program's version number and exit",
    )

    # set the argument parser to act as the parent
    parentparser = _utils.ArgumentParser(add_help=False)
    parentparser._optionals.title = "Verbosity options"
    parentparser.add_argument(
        "-v",
        "--verbose",
        action="count",
        default=0,
        help="increase verbose output",
    )
    parentparser.add_argument(
        "-s",
        "--silent",
        action="store_true",
        help="show only fatal errors",
    )

    # subparsers are dependent on which action is chosen
    subparsers = parser.add_subparsers(
        dest="mode", title="Actions",
        description="Select one of the following actions:")

    # Add the subparsers for each plot product
    for product, product_class in PRODUCTS.items():
        doc = product_class.__doc__ or ""
        subparser = subparsers.add_parser(
            product,
            help=doc.strip().split("\n", maxsplit=1)[0],
            parents=[parentparser],
            formatter_class=_utils.HelpFormatter,
        )
        product_class.init_cli(subparser)

    return parser


def parse_command_line(args: list[str] | None = None) -> Namespace:
    """Parse the command line arguments and return the parsed arguments."""
    parser = create_parser()
    return parser.parse_args(args=args)


# -- run -----------------------------

def main(args: list[str] | None = None) -> int:
    """Run gwpy-plot.

    Returns the relevant exit code, that can be passed to :func:`sys.exit`.
    """
    start = time.time()

    # parse the command line
    opts = parse_command_line(args=args)
    _utils.init_verbose_logging("gwpy", opts.verbose)
    logger.info("-- Welcome to gwpy-plot v%s --", __version__)

    # create a product object
    prod = PRODUCTS[opts.mode](opts)
    logger.debug("%s created", prod.action)

    # log how long it took us to get here
    setup_time = time.time() - start
    logger.debug("Setup time %.1fs", setup_time)

    # -- generate the plot
    prod.run()

    # overload the current namespace for interactive (i)python users
    if INTERACTIVE:
        # import pyplot so that user has access to it
        from matplotlib import pyplot as plt  # noqa: F401
        plot = prod.plot
        # pull raw data and plotted results from product for their use
        timeseries = prod.timeseries  # noqa: F841
        ax = plot.gca()  # noqa: F841
        print("Raw data is in 'timeseries', plot product is in 'prod'")  # noqa: T201

    run_time = time.time() - start
    logger.debug("Program run time: %.1fs", run_time)
    if prod.got_error:
        # Make sure when running batch they can test for error
        return 2
    return 0


if __name__ == "__main__":
    sys.exit(main())
